"""
base testing class to reduce code duplication
"""

import os
import shutil
import time
from typing import Union, Callable

import pytest

from remotemanager import Dataset
from remotemanager.utils import random_string


class BaseTestClass:
    datasets = []
    files = []
    kwarg_list = []
    fn_list = []

    @pytest.fixture(scope="function", autouse=True)
    def wrap(self):
        # print("Initialising test")
        self.setUp()

        yield  # test runs here

        print("Tearing down class")
        self.tearDown()

    def setUp(self):
        self.datasets: list[Dataset] = []
        self.files: list[str] = []

        self.kwarg_list = []
        self.fn_list = []

    def tearDown(self, wait_ds: bool = True):
        """Clean up"""

        if wait_ds:
            for ds in self.datasets:
                try:
                    ds.wait(0.1, 5)
                except RuntimeError:
                    pass

        for ds in self.datasets:
            self.destroy_dataset(ds)

        for file in self.files:
            try_remove(file)

        self.datasets = []
        self.files = []
        self.kwarg_list = []
        self.fn_list = []

    @property
    def ds(self) -> Dataset:
        return self.datasets[-1]

    def create_dataset(
        self, function, recreate: bool = False, **dataset_kwargs
    ) -> Dataset:
        """Generate a dataset for test usage"""
        print(f"creating dataset with function {function}")

        randstr = self.random_string()

        if "local_dir" not in dataset_kwargs:
            dataset_kwargs["local_dir"] = f"temp_local_{randstr}"

            ldir = dataset_kwargs["local_dir"]
            if os.path.exists(ldir):
                raise ValueError(f"local path {ldir} already exists!")

        if "remote_dir" not in dataset_kwargs:
            dataset_kwargs["remote_dir"] = f"temp_remote_{randstr}"

            rdir = dataset_kwargs["remote_dir"]
            if os.path.exists(rdir):
                raise ValueError(f"remote path {rdir} already exists!")

        if "skip" not in dataset_kwargs and not recreate:
            dataset_kwargs["skip"] = False
        if "name" not in dataset_kwargs and not recreate:
            dataset_kwargs["name"] = f"dataset_{randstr[:8]}"

        ds = Dataset(function=function, **dataset_kwargs)

        self.datasets.append(ds)
        print(f"created dataset with args {dataset_kwargs}")
        print(f"there are now {len(self.datasets)} datasets")

        self.kwarg_list.append(dataset_kwargs)
        self.fn_list.append(function)

        return ds

    def create_datasets(self, functions: list, **dataset_kwargs) -> list:
        randstr = self.random_string()

        link = dataset_kwargs.pop("link", True)

        if "local_dir" not in dataset_kwargs:
            dataset_kwargs["local_dir"] = f"temp_local_{randstr}"

        if "remote_dir" not in dataset_kwargs:
            dataset_kwargs["remote_dir"] = f"temp_remote_{randstr}"

        datasets = []
        for i, fn in enumerate(functions):
            dataset_kwargs["name"] = f"dataset_{i}_{fn.__name__}_{randstr}"
            ds = self.create_dataset(fn, **dataset_kwargs)

            if i > 0 and link:
                datasets[-1].set_downstream(ds)
            datasets.append(ds)

        return datasets

    @property
    def previous_ds_kwargs(self) -> dict:
        if len(self.kwarg_list) == 0:
            return {}
        return self.kwarg_list[-1]

    @property
    def previous_ds_fn(self) -> Union[Callable, None]:
        if len(self.fn_list) == 0:
            return None
        return self.fn_list[-1]

    def recreate_previous_dataset(self, **dataset_kwargs) -> Dataset:
        """
        Attempt to recreate the previously generated dataset

        .. note::
            Uses skip=True by default
        """
        kwargs = self.previous_ds_kwargs

        kwargs.update(dataset_kwargs)

        if "skip" not in dataset_kwargs:
            kwargs["skip"] = True

        print(f"(re) creating ds with kwargs {kwargs}")

        return Dataset(function=self.previous_ds_fn, **kwargs)

    def create_random_file(
        self,
        content: Union[str, None] = None,
        directory: Union[str, None] = None,
    ) -> str:
        if directory is not None and not os.path.exists(directory):
            os.makedirs(directory)

        filename = f"temp_{self.random_string()}"
        if directory is not None:
            filename = os.path.join(directory, filename)

        if content is not None:
            content = str(content) or ""
            with open(filename, encoding="utf8", mode="w+") as o:
                o.write(content)

        # store the filenames for later removal
        # Note that any copies will not be tracked
        self.files.append(filename)

        return filename

    def random_string(self, len: int = 16):
        return random_string(len=len)

    @staticmethod
    def destroy_dataset(dataset: Dataset):
        dataset.hard_reset(files_only=False, dependency_call=True, confirm=False)

        try_remove(dataset.dbfile)
        try_remove(dataset.local_dir)
        try_remove(dataset.remote_dir)

    def run_ds(self, interval=0.1, timeout=5, **kwargs) -> list:
        time.sleep(0.5)
        if len(self.datasets) == 0:
            return []

        self.ds.run(**kwargs)
        self.ds.wait(interval=interval, timeout=timeout)
        self.ds.fetch_results()

        return self.ds.results


def try_remove(f):
    try:
        os.remove(f)
    except IsADirectoryError:
        shutil.rmtree(f)
    except FileNotFoundError:
        pass



import logging
import os.path
import re
import xml.etree.ElementTree as ET
from typing import Union

from remotemanager.connection.computers.computer import Computer
from remotemanager.connection.computers.substitution import Substitution

logger = logging.getLogger(__name__)


class JUBETemplate(Computer):
    """Extends BaseComputer to provide compatibility with JUBE platforms"""

    def __init__(self, template: str, platform: str, **kwargs):
        if os.path.exists(template):
            with open(template) as o:
                template = o.read()

        self._platform_path = ""
        self.platform_path = platform

        super().__init__(template=template, **kwargs)

    @classmethod
    def from_repo(
        cls,
        path: str,
        branch: str = "develop",
        repo: str = "https://gitlab.com/max-centre/JUBE4MaX",
        platform_name: str = "platform.xml",
        template_name: str = "submit.job",
        local_dir: Union[None, str] = None,
        **kwargs,
    ):
        """
        Pulls a platform and template from a gitlab repository

        Files will be pulled from a combination of repo, branch and name

        e.g.
        from_repo(
            repo="https://gitlab.com/l_sim/remotemanager",
            branch="devel",
            name="tests/standard/foo"
        )

        This will look for the two links:
        https://gitlab.com/l_sim/remotemanager/-/raw/devel/tests/standard/foo/platform.xml
        https://gitlab.com/l_sim/remotemanager/-/raw/devel/tests/standard/foo/submit.job

        A general form of the links are:
        {repo}/-/raw/{branch}/{name}/{platform_name}
        {repo}/-/raw/{branch}/{name}/{template_name}
        """
        url = os.path.join("/", branch, path)
        url = f"{repo}/-/raw" + url

        if local_dir is None:
            local_dir = os.path.split(path.strip("/"))[-1]

            if local_dir == "" or local_dir is None:
                local_dir = os.getcwd()

        print(f"searching for {platform_name} & {template_name} at {url}")

        platform_path = os.path.join(local_dir, platform_name)
        cls.download_file(os.path.join(url, platform_name), platform_path)
        template_path = os.path.join(local_dir, template_name)
        cls.download_file(os.path.join(url, template_name), template_path)

        return cls(template=template_path, platform=platform_path, **kwargs)

    @property
    def platform_path(self) -> str:
        """Path to the platform file"""
        return self._platform_path

    @platform_path.setter
    def platform_path(self, path: str):
        if not os.path.exists(path):
            raise ValueError(f"Platform file {path} not found")

        self._platform_path = path

    def _extract_subs(self, link: bool = True) -> None:
        """
        Overrides BaseComputer sub extraction to instead parse from platform.xml
        """
        logger.debug("generating subs specified in platform.xml")
        tree = ET.parse(self.platform_path)
        root = tree.getroot()
        # platform.xml specifies "substitutesets" which translate what's
        # in the template to the actual parameter name. Apply these changes here.
        cache = {}
        for subsituteset in root.findall("substituteset"):
            for change in subsituteset.iter():
                if change.tag != "sub":
                    logger.debug("skip row %s", change)
                    continue
                target = change.attrib["source"]
                name = change.attrib["dest"].strip("$")

                cache[name] = {"target": target}

        for parameterset in root.findall("parameterset"):
            for parameter in parameterset.iter():
                if parameter.tag != "parameter":
                    continue

                name = parameter.attrib["name"]
                mode = parameter.attrib.get("mode", None)

                default = parameter.text

                if isinstance(default, str):
                    for tmp in cache:
                        if f"${tmp}" in default:
                            mode = "python"
                    default = default.strip()

                data = {"mode": mode, "default": default}
                if name in cache:
                    cache[name].update(data)
                else:
                    cache[name] = data
                # Subs require a target, create one if it does not exist
                target = cache[name].get("target", None)
                if target is None or target == "":
                    cache[name]["target"] = f"{name}"

        for name, args in cache.items():
            sub = Substitution(name=name, **args)
            self._subs[name] = sub

    def get_unevaluated_links(self, sub: Substitution) -> list:
        """
        Performs a regex search for unlinked JUBE $parameters within a value

        Args:
            sub: Substitution object to check
        """
        # only strings are valid targets
        if not isinstance(sub.value, str):
            return []
        # can lazily ignore any value not containing a $ char
        if "$" not in sub.value:
            return []
        # need to check for exact links using regex
        # this avoids bash $evaluations raising false positives
        cache = []
        for name in self.arguments:
            search = r"\$" + name + r"\b"
            if re.search(search, sub.value) is not None:
                cache.append(name)
        return cache

    def link_sub(self, sub: Substitution) -> None:
        """Performs recursive linking for target sub"""
        links = self.get_unevaluated_links(sub)

        # if the sub is fully linked, attempt an evaluation and return
        if len(links) == 0:
            logger.debug("%s has no children, evaluating and completing", sub.name)
            # set the temporary value, so these links do not permanently update values
            sub.temporary_value = attempt_eval(sub.value)
            return

        logger.debug("Performing recursive linking for sub %s", sub.name)
        # otherwise, we need to recursively continue
        value = sub.value  # store the value for treatment
        for name in links:
            logger.debug("\tevaluating child %s", name)
            obj = self._subs[name]  # get the actual associated object
            # if this sub child object has children of itself, recurse the call
            sub_links = self.get_unevaluated_links(obj)  # search for any children
            if len(sub_links) != 0:
                logger.debug("\tlinked to: %s, recursing...", sub_links)
                self.link_sub(obj)
            # recursion ensures that this child has a valid value, collect it
            update = str(obj.value)
            # search and replace an exact match
            search = r"\$" + name + r"\b"
            # search for the exact string, continuing if not found
            # this prevents $tasks matching $taskspernode, for example
            if re.search(search, value) is None:
                continue
            value = re.sub(search, update, value)
        # set the temporary value
        sub.temporary_value = attempt_eval(value)

    def _link_subs(self) -> None:
        """Ensures proper linking of JUBE substitutions"""
        for sub in self.sub_objects:
            self.link_sub(sub)


def attempt_eval(value: str):
    """Attempt to evaluate `value`"""
    try:
        return eval(value)
    except Exception:
        return value



"""Small container class for tracking runner insertion"""

from remotemanager.storage import SendableMixin


class SummaryInstance(SendableMixin):
    """
    Tracks a Runner insert instance

    Args:
        runner_id:
            uuid of runner
        mode:
            append mode: force, skip, etc.
        quiet:
            True if silent append
    """

    __slots__ = ["runner_id", "mode", "quiet"]

    def __init__(self, runner_id: str, mode: str, quiet: bool):
        self.runner_id = runner_id
        self.mode = mode
        self.quiet = quiet

    def __repr__(self):
        return f"{self.runner_id}: {self.mode}"



from remotemanager.dataset.repo import date_format


def create_run_function(
    submitter: str,
    manifest_filename: str,
    script_run: bool = False,
    add_docstring: bool = True,
) -> str:
    """
    Generates a submission function for submitter

    script_run should be used if the target function (or script) does not
    log its output to the manifest (as is the case with None as the function)

    $1 is the runner short_uuid
    $2 is the path to the jobscript
    $3 is the path to the error file
    $4 is the path to the result file

    Args:
        submitter: submitter to generate for
        manifest_filename: path to manifest file
        script_run: handle completed and failed status updates in the function
        add_docstring: adds docstring to function if True
    """
    template = f"""{{docstring}}
submit_job_{{submitter_cmd}} () {{
    echo "$(date -u +'{date_format}') [$1] submitted" >> \
$sourcedir/{{manifest_filename}}
    {{submission_section}}
}}"""

    submission_normal = f"""{{submitter}} $2 2> $3 ||
    echo "$(date -u +'{date_format}') [$1] failed" >> \
$sourcedir/{{manifest_filename}}
"""
    submission_script = f"""if {{submitter}} $2 > $4 2> $3 ; then
        echo "$(date -u +'{date_format}') [$1] completed" >> \
$sourcedir/{{manifest_filename}}
    else
        echo "$(date -u +'{date_format}') [$1] failed" >> \
$sourcedir/{{manifest_filename}}
    fi"""

    if script_run:
        template = template.replace("{submission_section}", submission_script)
    else:
        template = template.replace("{submission_section}", submission_normal)
    template = template.replace("{manifest_filename}", manifest_filename)
    template = template.replace("{submitter}", submitter)

    # strip any flags from the cmd
    template = template.replace("{submitter_cmd}", submitter.split(" ", maxsplit=1)[0])

    docstring = """# This function handles the running of jobs
# Arguments:
#   $1 is the runner short_uuid
#   $2 is the path to the jobscript
#   $3 is the path to the error file
#   $4 is the path to the result file"""

    if add_docstring:
        template = template.replace("{docstring}", docstring)
    else:
        template = template.replace("{docstring}", "")

    return template



import inspect
from typing import Union

import yaml

from remotemanager.connection.url import URL
from remotemanager.script.script import Script
from remotemanager.utils import get_version

package_collect = [
    "host",
    "user",
    "port",
    "timeout",
    "max_timeouts",
    "python",
    "submitter",
    "shell",
    "raise_errors",
    "sshpass_override",
    "cmd_history_depth",
    "landing_dir",
    "ssh_insert",
    "quiet_ssh",
    "shebang",
]


class Computer(Script, URL):
    """
    Combo class that allows for connection to a machine and generating jobscripts
    """

    def __init__(
        self,
        template: Union[str, None] = None,
        template_path: Union[str, None] = None,
        **kwargs,
    ):
        # super() behaves strangely with multiple inheritance
        # explicitly call the __init__ with self
        URL.__init__(self, **kwargs)
        Script.__init__(self, template=template, template_path=template_path, **kwargs)

    def pack(
        self,
        collect_values: bool = True,
        ignore_none: bool = True,
        prune_defaults: bool = True,
        *args,
        **kwargs,
    ) -> dict:
        """
        Package up this Computer to a dictionary that can be stored as a yaml file.

        A note on collection:
        The collected values are explicitly stated in `package_collect`.
        Automated collection is possible, but not feasible for "human readable" outputs

        __dict__ collects the internal variables
        So instead of `user`, you get the `_conn` dictionary

        Using dir() is an option, but collects far too many variables, and also
        has the possibility to accidentally call functions as it crawls the object

        Args:
            collect_values (bool):
                Also collect any stored values if True. Defaults to True.
            ignore_none (bool):
                skip any None values in serialisation. Defaults to True.
            prune_defaults (bool):
                Collects only non-default values if True. Defaults to True.

        Returns:
            dict:
                serialised output
        """
        # preface the data with the version info
        spec = {"remotemanager_version": get_version()}
        # collect info
        for item in package_collect:
            val = getattr(self, item, None)
            if val is None:
                continue
            spec[item] = val

        # cleanup for non-serialisable entities
        try:
            spec["verbose"] = self.verbose.level
        except AttributeError:
            spec["verbose"] = self.verbose

        # gather any non-private vals
        for k, v in self.__dict__.items():
            if k in spec:
                continue
            if ignore_none and v is None:
                continue
            if k not in self.args and not k.startswith("_"):
                spec[k] = v

        manual_defaults = {"landing_dir": "$HOME", "port": 22}
        if prune_defaults:
            # grab defaults for removal
            signature = inspect.signature(URL.__init__)
            for k, v in signature.parameters.items():
                if k in spec and spec.get(k, None) == v.default:
                    del spec[k]
            # special handler, for complex defaults
            for k, v in manual_defaults.items():
                if spec.get(k, None) == v:
                    del spec[k]

        spec["template"] = Script.pack(self, values=collect_values)

        return spec

    @classmethod
    def unpack(cls, data: dict, **kwargs) -> "Computer":
        data.update(kwargs)
        return cls(**data)

    def to_dict(self, *args, **kwargs) -> dict:
        if "collect_values" not in kwargs:
            kwargs["collect_values"] = False
        return self.pack(*args, **kwargs)

    @classmethod
    def from_dict(cls, data: dict, **kwargs) -> "Computer":
        return cls.unpack(data, **kwargs)

    def to_yaml(self, file: str = None, **kwargs) -> Union[None, str]:
        """
        Dump the computer to yaml. Returns the yaml content as a string if file is None
        """
        data = self.to_dict(**kwargs)

        # extract template so we can format it manually
        template = data.pop("template")
        template = "\n".join([f"  {line}" for line in template.split("\n")])

        content = yaml.dump(data, sort_keys=False)
        content += f"template: |\n{template}"

        if file is None:
            return content

        with open(file, mode="w+", encoding="utf8") as o:
            o.write(content)

    @classmethod
    def from_yaml(cls, file: str = None, data: str = None, **kwargs) -> "Computer":
        if file is None and data is None:
            raise ValueError(
                "Please provide a file path to file, or yaml content to data"
            )

        if file is not None:
            with open(file, mode="r", encoding="utf8") as o:
                data = yaml.safe_load(o)

        if isinstance(data, str):
            data = yaml.safe_load(data)

        return cls.from_dict(data=data, **kwargs)



import ast
import warnings

from IPython.core.magic import Magics, magics_class, cell_magic, needs_local_scope

import logging
from remotemanager import Dataset

logger = logging.getLogger(__name__)


@magics_class
class RCell(Magics):
    """
    Magic function that allows running an ipython cell on a remote machine
    with minimal lines of code.
    """

    @cell_magic
    @needs_local_scope
    def sanzu(self, line: str, cell: str, local_ns: dict) -> None:
        """
        Execute a decorators cell using an implicit remote Dataset

        Args:
            line:
                magic line, includes arguments for cell and Dataset
            cell:
                Cell contents
            local_ns:
                dict containing current notebook runtime attributes
        """

        def clean_line(inpstr: str) -> str:
            inpstr = inpstr.split("#")[0].strip()  # remove comments
            inpstr = inpstr.strip(",")  # remove any trailing commas
            return inpstr

        # split cell into sanzu args, function args and the actual cell content
        sanzu = [clean_line(line)]  # prime sanzu storage with initial line
        sargs = []
        spull = []
        cell_actual = []

        cell = cell.strip()

        if len(cell) == 0:
            raise ValueError("Cell has no content!")
        for line in cell.split("\n"):
            if line.startswith("%%"):
                if "sanzu" in line:
                    sanzu.append(clean_line(line.split("sanzu ")[-1]))
                elif "sargs" in line:
                    sargs.append(clean_line(line.split("sargs ")[1]))
                elif "spull" in line:
                    spull.append(clean_line(line.split("spull ")[1]))
            else:
                cell_actual.append(line)

        # clean the args, in case the cell is initialised with an empty `sanzu`
        sanzu = [line for line in sanzu if line != ""]

        logger.info("sanzu: %s", sanzu)
        logger.info("sargs: %s", sargs)
        logger.info("spull: %s", spull)
        logger.info("cell: %s", cell_actual)

        args = self.extract_in_scope(",".join(sanzu), local_ns)
        fargs = self.extract_in_scope(",".join(sargs), local_ns)
        fstr = self.parse_line(cell_actual, list(fargs), spull)

        logger.info("generated function string %s", fstr)
        logger.info("Dataset args: %s", args)
        logger.info("Function args: %s", fargs)

        run_skip = args.pop("skip", True)
        # Build the runner and run
        try:
            del local_ns["magic_dataset"]
        except KeyError:
            pass
        ds = Dataset(function=fstr, skip=run_skip, **args)
        ds.append_run(args=fargs)
        ds.run(skip=run_skip)

        if ds.run_cmd is not None and ds.run_cmd.stderr:
            raise RuntimeError(f"error detected in magic run: {ds.run_cmd.stderr}")
        ds.wait(1)
        ds.fetch_results()

        local_ns["magic_dataset"] = ds

        if ds.runners[0].is_failed:
            warnings.warn(
                "Sanzu encountered an exception, see below, "
                "or access magic_dataset.errors"
            )
            raise RuntimeError(ds.runners[0].error)

        for name in spull:
            logger.debug("looking for pull target %s", name)
            local_ns[name] = ds.results[0].get(name, None)

        return ds.results[0]

    def parse_line(self, cell: list, fargs: list, pull: list):
        """
        Generate a callable function from the remaining cell content

        Args:
            pull:
                (list): list of object names to add to return instead of final line
            cell:
                (list): cell content
            fargs:
                (list): function arguments

        Returns:
            (str): formatted function string
        """
        cell = [line for line in cell if line != ""]

        fstr = ["def __sanzu_fn("]
        if len(fargs) > 0:
            fstr[0] += ", ".join(fargs) + ", "
        fstr[0] += "):"

        fstr += [f"\t{line}" for line in cell]

        fstr = "\n".join(fstr)

        if len(pull) == 0:
            # Validate the function string and add a return
            fstr = self.parse_tree(fstr, add_return=True)
        else:
            fstr += (
                "\n\t_return_object = {}"
                f"\n\tfor _name in ({pull}):"
                "\n\t\t_return_object[_name] = locals().get(_name, None)"
                "\n\treturn _return_object"
            )
            fstr = self.parse_tree(fstr, add_return=False)

        return fstr

    def extract_in_scope(self, line: str, local_ns: dict):
        """
        This will convert a line to a dictionary of arguments.

        It is separated out because we are going to have to selectively
        populate the local scope with `local_ns`.
        """

        def kwargs_to_dict(**kwargs):
            return kwargs

        for k, v in local_ns.items():
            if k in line:
                locals()[k] = v

        return eval("kwargs_to_dict(" + line + ")")

    def parse_tree(self, fstr, add_return: bool = True):
        """
        This routine adds a return to the end of our generated function.
        """
        # current code, try to use this first
        if hasattr(ast, "unparse"):
            tree = ast.parse(fstr)
            node = tree.body[0]
            if add_return and isinstance(node.body[-1], ast.Expr):
                node.body[-1] = ast.Return(value=node.body[-1].value)

            return ast.unparse(tree)

        lines = fstr.split("\n")
        if add_return:
            # need to add a return, but check for cases to avoid
            rline = lines[-1]

            # check to see if the last line is indented
            # in most cases this should not be returned
            indent = round((len(rline) - len(rline.lstrip())) / 4)

            # generate the spacing from the indent
            spacer = "\t" * (indent + 1)

            # however, cells ending in an if statement can still return, check for this
            # try:
            #     penultimate = lines[-2].strip()
            # except IndexError:
            #     penultimate = ""
            #
            # skip_indent = indent > 0 and not penultimate.startswith("if")

            # ast.unparse does NOT have this behaviour, so comment it out
            skip_indent = indent > 0

            # need to check for an assignment
            # assume this is the case if there is only alphanumeric chars before the
            # first "=" character.
            # that is to say, foo="bar" has the letters f, o, o before
            # whereas func(a=10) contains the `(` char, and should be returned

            # this is inefficient but it's for compatibility reasons
            search = rline.replace(" ", "").strip()
            skip_assignment = False
            for char in search:
                # variable_names should not trip this
                if char in ("_", "."):
                    continue
                # loop has reached an = without exiting, don't add a return
                if char == "=":
                    skip_assignment = True
                    break
                # loop has reached a non-alphanumeric char before =, break
                if not char.isalnum():
                    break

            if skip_indent:
                pass
            elif skip_assignment:
                pass
            else:
                rline = f"{spacer}return {rline.strip()}"

                lines[-1] = rline

        return "\n".join(lines)



import hashlib
from typing import Any

from remotemanager.logging_utils.utils import format_iterable


def generate_uuid(string: Any) -> str:
    """
    Generates a UUID string from an input

    Args:
        string:
            input string
    Returns:
        (str) UUID
    """
    if not isinstance(string, str):
        string = format_iterable(string)
    h = hashlib.sha256()
    h.update(bytes(string, "utf-8"))

    return str(h.hexdigest())



from typing import Optional


def ask_confirm(
    msg: str = "Confirm action?",
    default: Optional[bool] = False,
):
    options = "[Y]/n" if default else "y/[N]"

    inp = input(f"{msg} {options}")

    if inp == "":
        inp = "y" if default else "n"

    if inp.lower().strip() == "y":
        return True
    return False



"""
Handles file transfer via `cp`
"""

import logging
from remotemanager.transport.transport import Transport

logger = logging.getLogger(__name__)


class cp(Transport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        logger.info("created new cp transport")

    def cmd(self, primary, secondary):
        cmd = "mkdir -p {secondary} ; cp -r --preserve {primary} {secondary}"
        base = cmd.format(primary=primary, secondary=secondary)
        logger.debug(f'returning formatted cmd: "{base}"')
        return base



"""
Main Dataset module

This is the primary class used by the user
"""

# pylint: disable=protected-access
import collections
import copy
import gc
import json
import logging
import os
import pathlib
import re
import shutil
import time
import warnings
from typing import Callable, List, Union
from zipfile import ZipFile

import yaml

from remotemanager.dataset.dataset_mixin import DetailMixin
import remotemanager.dataset.repo as repo
from remotemanager.dataset.runnerstates import RunnerState
import remotemanager.serialisation as sr
import remotemanager.transport as tp
from remotemanager.connection.cmd import CMD
from remotemanager.connection.computers.computer import Computer
from remotemanager.connection.computers.utils import try_value
from remotemanager.connection.url import URL
from remotemanager.dataset.dependency import Dependency
from remotemanager.dataset.files_mixin import ExtraFilesMixin
from remotemanager.dataset.function_generation import create_run_function
from remotemanager.dataset.lazy_append import LazyAppend
from remotemanager.dataset.runner import Runner, LOCALWINERROR
from remotemanager.dataset.summary_instance import SummaryInstance
from remotemanager.decorators.remotefunction import cached_functions
from remotemanager.logging_utils.decorate_verbose import make_verbose
from remotemanager.logging_utils.utils import format_iterable
from remotemanager.logging_utils.verbosity import Verbosity
from remotemanager.utils.confirm import ask_confirm
from remotemanager.script.script import Script
from remotemanager.storage import SendableMixin, TrackedFile
from remotemanager.storage.database import Database
from remotemanager.storage.function import Function
from remotemanager.utils import (
    get_version,
    ensure_list,
    ensure_filetype,
    check_dir_is_child,
)
from remotemanager.utils.uuid import generate_uuid

logger = logging.getLogger(__name__)


@make_verbose
class Dataset(SendableMixin, ExtraFilesMixin, DetailMixin):
    """
    Bulk holder for remote runs. The Dataset class handles anything regarding
    the runs as a group. Running, retrieving results, sending to remote, etc.

    Args:
        function (Callable, str, None):
            Function to run. Can either be the function object, source string or None
            If None, Runner will pass arguments to the `script` method
        url (URL):
            connection to remote (optional)
        transport (tp.transport.Transport):
            transport system to use, if a specific is required. Defaults to
            transport.rsync
        serialiser (serial.serial):
            serialisation system to use, if a specific is required. Defaults
            to serial.serialjson
        script (str):
            callscript required to run the jobs in this dataset
        submitter (str):
            command to exec any scripts with. Defaults to "bash"
        name (str):
            optional name for this dataset. Will be used for runscripts
        extra_files_send(list, str):
            extra files to send with this run
        extra_files_recv(list, str):
            extra files to retrieve with this run
        skip (bool):
            skip dataset creation if possible. Defaults True
        extra:
            extra text to insert into the runner jobscripts
        global_run_args:
            any further (unchanging) arguments to be passed to the runner(s)

    Attributes:
        default_url (URL):
            a default url can be assigned to all Datasets.
    """

    _do_not_package = ["_database", "_url"]

    _manifest_file = "archive_manifest.txt"

    default_url = None

    # DEV NOTE: arguments must be None for computer-url override to function
    def __init__(
        self,
        function: Union[Callable, str, None],
        url: URL = None,
        dbfile: str = None,
        transport: tp.transport.Transport = None,
        serialiser: sr.serial = None,
        script: str = None,
        shebang: str = None,
        name: str = None,
        extra_files_send: list = None,
        extra_files_recv: list = None,
        verbose: Union[None, int, bool, Verbosity] = None,
        run_summary_limit: int = 25,
        add_newline: bool = True,
        skip: bool = True,
        extra: str = None,
        **global_run_args,
    ):
        self.verbose = verbose

        self.verbose.print("Dataset initialised", 2)
        logger.info("dataset initialised")

        if isinstance(function, Script):
            self._function = function
        elif function is not None:
            self._function = Function(function)
        else:
            logger.info("creating a dataset with no function")
            self._function = None
            if not isinstance(url, Computer):
                logger.warning(
                    "Function is None, but the URL is not a Computer subclass"
                )
                print(
                    "Warning! The current url is "
                    "not a subclass of Computer, "
                    "the dataset may not function as expected."
                )

        self.run_args = copy.deepcopy(Runner._defaults)
        # sanitise paths
        self.run_args = self.sanitise_run_arg_paths(self.run_args)

        self.run_args.update(global_run_args)
        self._global_run_extra = extra

        # dataset uuid is equal to Function uuid for now
        self._name = None
        self.name = name or "dataset"
        try:
            source_uuid = self._function.uuid
        except AttributeError:
            source_uuid = self.url.uuid
        self._uuid = generate_uuid(source_uuid + self.name)
        logger.info("uuid is %s", self.uuid)

        self._extra_files = {
            "send": (
                ensure_list(extra_files_send) if extra_files_send is not None else []
            ),
            "recv": (
                ensure_list(extra_files_recv) if extra_files_recv is not None else []
            ),
        }
        self._add_newline = add_newline

        self._last_run = -1

        self._url = None
        self._transport = None
        self._computer = False
        self._serialiser = None
        self._dependency = None
        self._do_not_recurse = False

        self._dbfile_override = dbfile
        self._repo_prefix = None

        if "dataset" in self.name.lower():
            dbfile_base = f"{self.name}-{self.short_uuid}"
        else:
            dbfile_base = f"dataset-{self.name}-{self.short_uuid}"

        self._dbfile = ensure_filetype(dbfile_base, "yaml")

        self.main_dir_env = f"DATASET_{self.short_uuid}_MAIN_DIR"

        self.argfile = f"args-{self.name}-{self.short_uuid}"

        if url is None:
            self.url = self.default_url
        else:
            self.url = url
        self.transport = transport
        self.serialiser = serialiser

        if "submitter" in global_run_args:
            self.submitter = global_run_args["submitter"]
        if "shell" in global_run_args:
            self.shell = global_run_args["shell"]

        self._script = script or ""
        # Dataset shebang takes priority _only_ if manually set
        if shebang is not None:
            self.shebang = shebang

        if not skip:
            try:
                os.remove(self.dbfile)
                logger.info("deleted database file %s", self.dbfile)
            except FileNotFoundError:
                pass

        self._run_cmd = None
        self._fresh_dataset = False
        if skip and os.path.isfile(self.dbfile):
            self._create_from_db()
        else:
            self._create_fresh()

        self._append_session = 0
        self._run_summary_limit = run_summary_limit

        logger.info("Dataset %s init complete)", self.name)

    def _create_from_db(self) -> None:
        logger.info("unpacking database from %s", self.dbfile)
        self.verbose.print(f"Unpacking Dataset from {self.dbfile}", 2)

        # create a "temporary" database from the found file
        self._database = Database(self.dbfile)
        old_uuid = self.database.stored_uuid  # get uuid by first key
        logger.info("unpacked uuid is %s", old_uuid)
        if old_uuid != self.uuid:
            logger.debug("current uuid is %s", self.uuid)
            dst = f"{self.dbfile}.old"

            dst = self.database.backup(dst)
            msg = (
                f"new and old UUID mismatch (did something change?)\n"
                f"Creating a fresh dataset and backing up the old dbfile at {dst}."
                f"\nUse Dataset.from_file('{dst}') to recover the old dataset."
            )
            logger.warning(msg)
            print(msg)
            self._create_fresh()
            return
        # update it with any new values
        self.update_db()
        # unpack from here to retrieve
        payload = self.database._storage[self.uuid]
        self.inject_payload(payload)

    def _create_fresh(self) -> None:
        logger.info("No database file found, creating anew")
        self.verbose.print(f"Creating a fresh Dataset w/ database at {self.dbfile}", 2)
        self._runs = collections.OrderedDict()
        self._uuids = []
        self._results = []

        # database property creates the database if it does not exist
        self.database._storage = {}
        self.update_db()
        self._fresh_dataset = True

    def __hash__(self) -> int:
        return hash(self.uuid)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.uuid == other.uuid

    def __repr__(self):
        return f"dataset-{self.name}-{self.short_uuid}"

    @classmethod
    def recreate(cls, *args, raise_if_not_found: bool = True, **kwargs):
        """
        Attempts to extract a dataset matching the given args from the python
        garbage collection interface

        Args:
            raise_if_not_found (bool):
                raise ValueError if the Dataset was not found
            *args:
                args as passed to Dataset
            **kwargs:
                keyword args as passed to Dataset
        Returns:
            Dataset
        """

        for obj in gc.get_objects():
            if isinstance(obj, cls):
                tmp = cls(*args, **kwargs)
                if obj == tmp:
                    print("returning stored obj")
                    return obj

        if raise_if_not_found:
            raise ValueError("Dataset with args not found!")

        return cls(*args, **kwargs)

    @classmethod
    def from_file(cls, file: str, url: URL = None):
        """
        Alias for Dataset.unpack(file=...)

        Args:
            file (str):
                Dataset dbfile
            url (URL):
                the URL to apply to this Dataset

        Returns:
            (Dataset): unpacked Dataset
        """
        ds = Dataset.unpack(file=file)

        if ds is not None:
            ds.url = url
        else:
            print("Warning! Creating Dataset from file with a default URL (localhost)")

        return ds

    @property
    def database(self) -> Database:
        """
        Access to the stored database object.
        Creates a connection if none exist.

        Returns (Database):
            Database
        """
        if self.is_missing("_database"):
            logger.debug("Database missing, regenerating from file %s", self.dbfile)
            payload = self.pack()
            payload[Database._versionkey] = get_version()
            with open(self.dbfile, "w+", encoding="utf8") as o:
                yaml.dump(payload, o)

            self._database = Database(file=self.dbfile)
        return self._database

    @property
    def dbfile(self) -> str:
        """
        Name of the database file
        """
        if self._dbfile_override is not None:
            return ensure_filetype(self._dbfile_override, "yaml")
        return self._dbfile

    @dbfile.setter
    def dbfile(self, file: str) -> None:
        """
        Moves the stored database to `file`.

        Args:
            file:
                new file path
        """
        logger.debug("updating dbfile to %s", file)
        self.update_db()  # make sure the database is valid
        # make sure new path is valid
        path = ensure_filetype(file, "yaml")
        # move the old database to the new location
        shutil.move(self.dbfile, path)
        logger.debug("moved %s to %s", self.dbfile, path)
        self._dbfile = path  # update internal path

    def sanitise_run_arg_paths(self, run_args: dict) -> dict:
        """
        Checks for issues in the paths within the given run_args
        """
        if "remote_dir" in run_args:
            run_args["remote_dir"] = self.sanitise_path(run_args["remote_dir"])
        if "run_dir" in run_args:
            run_args["run_dir"] = self.sanitise_path(run_args["run_dir"])
        if "local_dir" in run_args:
            run_args["local_dir"] = self.sanitise_path(run_args["local_dir"])

        return run_args

    @staticmethod
    def sanitise_path(path) -> str:
        """
        Ensures a clean unix-type path
        """
        path = str(pathlib.PureWindowsPath(path).as_posix())

        if " " in path:
            raise ValueError(f"Space character detected in path {path}")

        return path

    @property
    def remote_dir(self) -> Union[str, bool]:
        """
        Accesses the remote_dir property from the run args. Tries to fall back
        on run_dir if not found, then returns default as a last resort.
        """
        return self.global_run_args.get("remote_dir", Runner._defaults["remote_dir"])

    @remote_dir.setter
    def remote_dir(self, path):
        self.run_args["remote_dir"] = self.sanitise_path(path)

        self.update_db()

    @property
    def run_dir(self) -> Union[str, bool]:
        """
        Accesses the remote_dir property from the run args. Tries to fall back
        on run_dir if not found, then returns default as a last resort.
        """
        return self.global_run_args.get("run_dir", None)

    @run_dir.setter
    def run_dir(self, path):
        self.run_args["run_dir"] = self.sanitise_path(path)

        self.update_db()

    @property
    def run_path(self) -> Union[str, bool]:
        """
        Accesses the remote_dir property from the run args. Tries to fall back
        on run_dir if not found, then returns default as a last resort.
        """
        if self.run_dir is not None:
            return os.path.join(self.remote_dir, self.run_dir)
        return self.remote_dir

    @property
    def local_dir(self) -> str:
        """
        Accesses the local_dir property from the run args. Returns default if
        not found.
        """
        return self.global_run_args.get("local_dir", Runner._defaults["local_dir"])

    @local_dir.setter
    def local_dir(self, path):
        self.run_args["local_dir"] = self.sanitise_path(path)

        self.update_db()

    @property
    def repo_prefix(self) -> str:
        """override for repo names and manifest file in a dependency situation"""
        if self._repo_prefix is None:
            return f"{self.name}-{self.short_uuid}"
        return self._repo_prefix

    @property
    def repofile(self) -> TrackedFile:
        """Returns the TrackedFile instance responsible for the repository"""
        fname = f"{self.repo_prefix}-repo.py"
        return TrackedFile(self.local_dir, self.remote_dir, fname)

    @property
    def bash_repo(self) -> TrackedFile:
        """Returns the TrackedFile instance responsible for the repository"""
        fname = f"{self.repo_prefix}-repo.sh"
        return TrackedFile(self.local_dir, self.remote_dir, fname)

    @property
    def master_script(self) -> TrackedFile:
        """Returns the TrackedFile instance responsible for the master script"""
        fname = f"{self.name}-{self.short_uuid}-master.sh"
        return TrackedFile(self.local_dir, self.remote_dir, fname)

    @property
    def manifest_log(self) -> TrackedFile:
        """Returns the TrackedFile instance responsible for the manifest"""
        fname = f"{self.repo_prefix}.manifest"
        return TrackedFile(self.local_dir, self.remote_dir, fname)

    @property
    def global_run_args(self) -> dict:
        """Returns the toplevel global run args"""
        return self.run_args

    def set_run_arg(self, key: str, val):
        """
        Set a single run arg `key` to `val`

        Args:
            key:
                name to set
            val:
                value to set to

        Returns:
            None
        """
        self.run_args[key] = val

    def set_run_args(self, keys: list, vals: list):
        """
        Set a list of `keys` to `vals

        .. note::
            List lengths must be the same

        Args:
            keys:
                list of keys to set
            vals:
                list of vals to set to

        Returns:
            None
        """
        keys = ensure_list(keys)
        vals = ensure_list(vals)

        if len(keys) != len(vals):
            raise ValueError(
                f"number of keys ({len(keys)}) != number of vals ({len(vals)}"
            )

        for key, val in zip(keys, vals):
            self.run_args[key] = val

    def update_run_args(self, d: dict):
        """
        Update current global run args with a dictionary `d`

        Args:
            d:
                dict of new args

        Returns:
            None
        """
        self.run_args.update(d)

    @property
    def do_not_recurse(self) -> bool:
        """Internal function used for blocking recursion in dependency calls"""
        self._do_not_recurse = False
        return True

    @property
    def dependency(self) -> Union[Dependency, None]:
        """Returns the stored dependency"""
        return self._dependency

    @property
    def is_child(self) -> bool:
        """Returns True if this dataset is a child, False otherwise"""
        if self.dependency is None:
            return False
        return self.short_uuid in self.dependency._children

    @property
    def is_parent(self) -> bool:
        """Returns True if this dataset is a parent, False otherwise"""
        if self.dependency is None:
            return False
        return self.short_uuid in self.dependency._parents

    def _mirror_dependency(self, dataset) -> None:
        """Ensures bi-directional dependency links"""
        logger.info("connecting with dataset %s", dataset)
        if dataset.dependency is not None:
            logger.info("target has dependency, joining")
            self._dependency = dataset.dependency
        elif self.dependency is not None:
            logger.info("self has dependency, joining")
            dataset._dependency = self._dependency
        else:
            logger.info("creating a dependency and entering")
            self._dependency = Dependency()
            dataset._dependency = self.dependency

    def set_downstream(self, dataset) -> None:
        """Add a child to this dataset"""
        self._mirror_dependency(dataset)
        dataset._repo_prefix = self.repo_prefix

        self.dependency.add_edge(self, dataset)

        if not dataset.do_not_recurse:
            dataset._do_not_recurse = True
            dataset.set_upstream(self)
        self.update_db()

    def set_upstream(self, dataset) -> None:
        """Add a parent to this dataset"""
        self._mirror_dependency(dataset)
        self._repo_prefix = dataset.repo_prefix

        self.dependency.add_edge(dataset, self)

        if not dataset.do_not_recurse:
            dataset._do_not_recurse = True
            dataset.set_downstream(self)
        self.update_db()

    def pack(self, file: str = None, **kwargs) -> Union[dict, None]:
        """
        Override for the SendableMixin.pack() method, ensuring the dataset is
        always below a ``uuid``

        Args:
            **kwargs:
                Any arguments to be passed onwards to the SendableMixin.pack()

        Returns:
            (dict) packing result
        """
        if len(kwargs) == 0:
            logger.info("Dataset override pack called")
        else:
            logger.info("Data override pack called with run_args")
            logger.info("%s", format_iterable(kwargs))
        data = super().pack(uuid=self._uuid, **kwargs)

        if file is not None:
            print(f"dumping payload to {file}")
            with open(file=file, mode="w+", encoding="utf8") as o:
                yaml.dump(data, o)
            return None

        return data

    def update_db(self, dependency_call: bool = False) -> None:
        """Force updates the database"""
        if self.dependency is not None and not dependency_call:
            return self.dependency.update_db()

        self.database.update(self.pack())

    def set_run_option(self, key: str, val) -> None:
        """
        Update a global run option `key` with value `val`

        Args:
            key (str):
                option to be updated
            val:
                value to set
        """
        warnings.warn("set_run_option is deprecated, use set_run_arg instead")
        self.run_args[key] = val

    def append_run(
        self,
        args: dict = None,
        arguments: dict = None,
        name: str = None,
        extra_files_send: Union[list, str, None] = None,
        extra_files_recv: Union[list, str, None] = None,
        dependency_call: bool = False,
        verbose: int = None,
        quiet: bool = False,
        skip: bool = True,
        force: bool = False,
        lazy: bool = False,
        chain_run_args: bool = True,
        extra: str = None,
        return_runner: bool = False,
        **run_args,
    ):
        """
        Serialise arguments for later runner construction

        Args:
            args (dict):
                dictionary of arguments to be unpacked
            arguments (dict):
                alias for args
            name (str):
                 append a runner under this name
            extra_files_send (list, str):
                extra files to send with this run
            extra_files_recv (list, str):
                extra files to retrieve with this run
            dependency_call (bool):
                True if called via the dependency handler
            verbose (int, Verbose, None):
                verbose level for this runner (defaults to Dataset level)
            quiet (bool):
                disable printing for this append if True
            skip (bool):
                ignores checks for an existing runner if set to False
            force (bool):
                always appends if True
            lazy (bool):
                performs a "lazy" append if True, skipping the dataset update. You MUST
                call ds.finish_append() after you are done appending to avoid strange
                behaviours
            chain_run_args (bool):
                for dependency runs, will not propagate run_args to other datasets in
                the chain if False (defaults True)
            extra:
                extra string to add to this runner
            return_runner:
                returns the appened (or matching) runner if True
            run_args:
                any extra arguments to pass to runner
        """
        if quiet or lazy:
            verbose = Verbosity(0)
        elif verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose

        logger.debug("#### Dataset append_run called")
        if args is None and arguments is not None:
            args = arguments

        if self.dependency is not None and not dependency_call:
            return self.dependency.append_run(
                caller=self,
                chain_run_args=chain_run_args,
                args=args,
                name=name,
                extra_files_send=extra_files_send,  # noqa: E251
                extra_files_recv=extra_files_recv,  # noqa: E251
                verbose=verbose,
                quiet=quiet,
                skip=skip,
                force=force,
                lazy=lazy,
                run_args=run_args,
                extra=extra,
                return_runner=return_runner,
            )

        extra_files_send = ensure_list(extra_files_send) + self._extra_files["send"]
        extra_files_recv = ensure_list(extra_files_recv) + self._extra_files["recv"]

        rnum = len(self.runners)
        if name is not None:
            run_args["name"] = name
            r_id = name

            if name in self.runner_dict:
                msg = f"{self} overwriting already existing runner {r_id}"
                logger.warning(msg)
                verbose.print(msg, 1)

        else:
            r_id = f"runner-{rnum}"

        tmp = Runner(
            arguments=args,
            dbfile=self.dbfile,
            parent=self,
            self_id=r_id,
            extra_files_send=extra_files_send,
            extra_files_recv=extra_files_recv,
            verbose=verbose,
            extra=extra,
            **run_args,
        )

        tmp.result_extension = self.serialiser.extension

        tmp = self.insert_runner(
            runner=tmp,
            skip=skip,
            force=force,
            lazy=lazy,
            quiet=quiet,
            return_runner=return_runner,
        )

        if return_runner:
            return tmp

    def insert_runner(
        self,
        runner: Runner,
        skip: bool = True,
        force: bool = False,
        lazy: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
        quiet: bool = False,
        return_runner: bool = False,
    ) -> Union[None, Runner]:
        """
        Internal runner insertion.

        Args:
            runner: Runner object to insert
            skip: don't insert if it exists
            force: force inserts
            lazy: Attempts a lazy append if True (does not update DB)
            verbose: Verbosity level for this runner
            quiet: inserts runner quietly if True
            return_runner: Returns the runner object if True
        Returns:
            None or Runner
        """
        if quiet or lazy:
            verbose = Verbosity(0)
        elif verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose

        if runner.parent != self:
            logger.info("inserting runner from another dataset, overriding parent")

            runner._parent = self
            runner._parent_uuid = self.uuid

        def append_to_log(r_id, mode, quiet):
            session = f"append session {self._append_session}"
            insert = SummaryInstance(r_id, mode, quiet)
            try:
                self._append_log[session].append(insert)
            except KeyError:
                self._append_log[session] = [insert]

        r_id = runner.id

        if force or not skip:
            self._runs[r_id] = runner
            # regenerate a ``uuid`` so this runner can be properly tracked
            runner._generate_uuid({"r_id": r_id})

            self._uuids.append(runner.uuid)
            msg = f"force appended run {runner.name}"
            logger.info(msg)
            append_to_log(r_id, "forced", quiet)
            if verbose:
                verbose.print(msg, level=1)
        elif runner.uuid not in self._uuids:
            self._runs[r_id] = runner
            self._uuids.append(runner.uuid)
            msg = f"appended run {runner.name}"
            logger.info(msg)
            append_to_log(r_id, "appended", quiet)
            if verbose:
                verbose.print(msg, level=1)
        else:
            runner = self.get_runner(runner.uuid)
            msg = f"runner {runner.name} already exists"
            logger.info(msg)
            append_to_log(r_id, "skipped", quiet)
            if verbose:
                verbose.print(msg, level=1)

        if not lazy:
            self.finish_append(print_summary=False)

        if return_runner:
            return runner
        return None

    def finish_append(
        self,
        dependency_call: bool = False,
        print_summary: bool = True,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> None:
        """
        Completes the append process by updating the database, and printing a summary
        if necessary

        Args:
            dependency_call:
                Will not attempt to relay to a dependency
                if True (called by dependency)
            print_summary:
                Prints a summary if True
            verbose:
                verbosity level for this call
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        if print_summary and len(self._append_log) != 0:
            self._append_session += 1
            session = list(self._append_log.keys())[-1]
            summary = {}
            print_flag = False
            for instance in self._append_log[session]:
                if not instance.quiet:
                    print_flag = True
                try:
                    summary[instance.mode] += 1
                except KeyError:
                    summary[instance.mode] = 1
            msg = [f"Of {sum(summary.values())} appends:"]
            for mode, count in summary.items():
                msg.append(f"{count} {mode}")

            sessionlog = " ".join(msg)
            logger.info(sessionlog)
            if print_flag:
                verbose.print(sessionlog, 1)
                verbose.print("See get_append_log for more info", 1)

        logger.info("finishing append")
        if self.dependency is not None and not dependency_call:
            self.dependency.finish_append()
            return
        self.update_db()

    def lazy_append(self) -> LazyAppend:
        """Access a LazyAppend object, which handles the append finalisation"""
        return LazyAppend(self)

    def copy_runners(self, dataset: "Dataset") -> None:
        """Copy the runners from dataset over to this dataset"""
        for runner in dataset.runners:
            copied_runner = copy.deepcopy(runner)
            if runner.is_finished:
                copied_runner.set_state("copied", value=runner.state.value)
            else:
                copied_runner.set_state("copied", 0)

            self.insert_runner(copied_runner, lazy=True)
        self.finish_append(verbose=False)

    def remove_run(
        self,
        ident: Union[int, str, dict],
        dependency_call: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> bool:
        """
        Remove a runner with the given identifier. Search methods are identical
        get_runner(id)

        Args:
            ident:
                identifier
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            verbose:
                local verbose level

        Returns:
            (bool): True if succeeded
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        if not dependency_call and self.dependency is not None:
            return self.dependency.remove_run(ident)

        runner = self.get_runner(ident, dependency_call, verbose=0)

        if runner is None:
            logger.info(
                "could not find runner to remove",
            )
            return False

        del self._runs[runner.id]
        self._uuids.remove(runner.uuid)

        msg = f"removed runner {runner}"
        logger.info(msg)
        verbose.print(msg, level=1)

        # need to override attribute first, as updating can only add
        self.database._storage[self.uuid]["_runs"] = {}
        self.update_db()

        return True

    def get_runner(
        self,
        ident: Union[int, str, dict],
        dependency_call: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> Union[Runner, None]:
        """
        Collect a runner with the given identifier. Depending on the type of
        arg passed, there are different search methods:

        - int: the runners[ident] of the runner to remove
        - str: searches for a runner with the matching uuid
        - dict: attempts to find a runner with matching args

        Args:
            ident:
                identifier
            dependency_call (bool):
                used by the dependencies, runners cannot be removed via uuid in this
                case, as the uuids will not match between datasets

        Returns:
            (Runner): collected Runner, None if not available
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        verbose.print("Searching for runner", level=2, end="... ")

        def get_by_id(ident):
            logger.info("getting runner by id %s", ident)
            verbose.print(f"by id {ident}", level=2, end=" ")
            try:
                key = list(self.runner_dict.keys())[ident]
                return self.runner_dict[key]
            except IndexError:
                return

        def get_by_str(ident):
            logger.info('getting runner by string "%s"', ident)
            verbose.print(f"by string {ident}", level=2, end=" ")

            if ident.lower() in self.runner_dict:
                return self.runner_dict[ident.lower()]

            if dependency_call:
                raise RuntimeError(
                    "Runners within a dependency cannot be removed by uuid\n"
                    f"If trying to remove by name, there may be an error ({ident})."
                )
            # assume uuid at this point, search first by matching the first 8 chars
            # short_uuid, then confirming with the full, if given
            if len(ident) == 64:
                logger.info(
                    "full uuid)",
                )
                for runner in self.runners:
                    if runner.uuid == ident:
                        return runner
            elif len(ident) == 8:
                logger.info(
                    "short uuid)",
                )
                for runner in self.runners:
                    if runner.short_uuid == ident:
                        return runner

        def get_by_dict(ident):
            verbose.print(f"by args {ident}", level=2, end=" ")
            logger.info("getting runner by args %s", ident)
            for r in self.runners:
                if format_iterable(r.args) == format_iterable(ident):
                    return r

        dispatch = {int: get_by_id, str: get_by_str, dict: get_by_dict}

        runner = dispatch.get(type(ident))(ident)
        verbose.print(runner, level=2)
        return runner

    def wipe_runs(self, dependency_call: bool = False, confirm: bool = True) -> None:
        """
        Removes all runners

        Args:
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            confirm (bool):
                Asks for confirmation if True
        """
        if confirm and not ask_confirm("Remove all Runners?"):
            return

        if not dependency_call and self.dependency is not None:
            self.dependency.clear_runs()
            return

        logger.info("wiping all runners and updating the db")

        self._uuids = []
        self._runs = {}

        self.database._storage[self.uuid]["_runs"] = {}
        self.update_db()

    def reset_runs(
        self, wipe: bool = False, dependency_call: bool = False, confirm: bool = True
    ) -> None:
        """
        Remove any results from the stored runners and attempt to delete their
        result files if `wipe=True`

        .. warning::
            This is a potentially destructive action, be careful with this
            method

        Args:
            wipe:
                Additionally deletes the local files if True. Default False
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            confirm (bool):
                Asks for confirmation if True
        """
        if confirm and not ask_confirm("Reset all Runner results?"):
            return

        if not dependency_call and self.dependency is not None:
            self.dependency.clear_results(wipe)
            return

        logger.info("clearing results")
        for runner in self.runners:
            runner.clear_result(wipe)

    def collect_files(
        self,
        remote_check: bool,
        results_only: bool = False,
        extra_files_send: bool = True,
    ) -> list:
        """
        Collect created files

        Args:
            remote_check:
                search for remote paths if True
            results_only:
                only collect files that are returned from a run such as Results and
                extra_files_recv if True
            extra_files_send:
                collects extra_files_send if True

        Returns:
            list of filepaths
        """
        target = "remote" if remote_check else "local"

        targets = []

        if not results_only:
            targets = [
                getattr(self.master_script, target),
                getattr(self.repofile, target),
                getattr(self.bash_repo, target),
                getattr(self.manifest_log, target),
            ]

        # grab all runner files
        for runner in self.runners:
            # start with constants
            targets += [
                getattr(runner.resultfile, target),
                getattr(runner.errorfile, target),
            ]
            # add the jobscript and runfile if we want all files
            if not results_only:
                targets += [
                    getattr(runner.jobscript, target),
                    getattr(runner.runfile, target),
                ]

                # need extra files, within their remote/local dir
                # also needs to be a copy of this list to prevent
                # remote check contamination
                extras = [f for f in runner.extra_files_recv]
                if extra_files_send:
                    for file in runner.extra_files_send:
                        extras.append(file)

                for file in extras:
                    targets.append(getattr(file, target))

        # minimize length
        targets = list(set(targets))

        return targets

    def wipe_local(
        self,
        files_only: bool = True,
        dry_run: bool = False,
        dependency_call: bool = False,
        confirm: bool = True,
    ) -> None:
        """
        Clear out the local directory

        Args:
            files_only (bool):
                delete individual files instead of whole folders (preserves
                extra files)
            dry_run (bool):
                print targets and exit
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            confirm (bool):
                Asks for confirmation if True
        """
        if confirm and not ask_confirm("Wipe the local directory?"):
            return

        if not dependency_call and self.dependency is not None:
            return self.dependency.wipe_local(files_only)

        if not files_only and not check_dir_is_child(os.getcwd(), self.local_dir):
            raise RuntimeError(
                f"local dir {self.local_dir} is not a child directory, "
                f"deleting could have catastrophic effects"
            )

        logger.debug("wiping local")

        if not files_only:
            targets = [self.local_dir]
            for runner in self.runners:
                if runner.local_dir not in targets:
                    targets.append(runner.local_dir)

            logger.debug("locals: %s", format_iterable(targets))
            if dry_run:
                for local in targets:
                    print(f"targeting local {local} for wipe")

            for local in targets:
                try:
                    shutil.rmtree(local)
                    logger.debug("%s removed)", local)
                except FileNotFoundError:
                    logger.debug("%s not found", local)

        else:
            logger.debug("file only wipe")
            targets = self.collect_files(remote_check=False, extra_files_send=False)

            logger.info("targets for wipe:|%s", format_iterable(targets))

            if dry_run:
                for local in targets:
                    print(f"targeting local {local} for wipe")

            for path in targets:
                try:
                    if "*" in path:
                        logger.debug("skipping wildcard: %s", path)
                        continue
                    os.remove(path)
                    logger.debug("removed: %s", path)
                except FileNotFoundError:
                    logger.debug("not found: %s", path)

    def wipe_remote(
        self,
        files_only: bool = True,
        dry_run: bool = False,
        dependency_call: bool = False,
        confirm: bool = True,
    ) -> None:
        """
        Clear out the remote directory (including run dir)

        Args:
            files_only (bool):
                delete individual files instead of whole folders (preserves
                extra files)
            dry_run (bool):
                print targets and exit
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            confirm (bool):
                Asks for confirmation if True
        """
        if confirm and not ask_confirm("Wipe the remote directory?"):
            return
        logger.debug("wiping remote")

        if not dependency_call and self.dependency is not None:
            self.dependency.wipe_remote(files_only)
            return

        if not files_only:
            remotes = [self.remote_dir]
            for runner in self.runners:
                if runner.remote_dir not in remotes:
                    remotes.append(runner.remote_dir)
                if runner.run_path not in remotes:
                    remotes.append(runner.run_path)

            logger.debug("remotes: %s", format_iterable(remotes))
            if dry_run:
                for remote in remotes:
                    print(f"targeting remote {remote} for wipe")

            remotestr = ",".join(remotes)
            if len(remotes) > 1:
                cmd = f"rm -rf {{{remotestr}}}"
            else:
                cmd = f"rm -rf {remotestr}"

            self.url.cmd(cmd)

        else:
            logger.debug("file only wipe")
            targets = self.collect_files(remote_check=True)

            logger.info("targets for wipe:|%s", format_iterable(targets))

            # skip any wildcards
            targets = [t for t in targets if "*" not in t]

            cmd = ",".join(targets)
            cmd = f"rm -rf {{{cmd}}}"

            if dry_run:
                for remote in targets:
                    print(f"targeting remote {remote} for wipe")
                return

            self.url.cmd(cmd)

    def hard_reset(
        self,
        files_only: bool = True,
        dry_run: bool = False,
        dependency_call: bool = False,
        confirm: bool = True,
    ) -> None:
        """
        Hard reset the dataset, including wiping local and remote folders

        Args:
            files_only (bool):
                delete individual files instead of whole folders (preserves
                extra files)
            dry_run (bool):
                print targets and exit
            dependency_call (bool):
                used by any dependencies that exist, prevents recursion
            confirm (bool):
                Asks for confirmation if True
        """
        if confirm and not ask_confirm("Perform a hard reset?"):
            return

        if not dependency_call and self.dependency is not None:
            self.dependency.hard_reset(files_only)
            return

        self._dependency = None
        self.wipe_local(
            files_only, dry_run, dependency_call=dependency_call, confirm=False
        )
        self.wipe_remote(
            files_only, dry_run, dependency_call=dependency_call, confirm=False
        )
        self.wipe_runs(dependency_call=dependency_call, confirm=False)

        try:
            os.remove(self.dbfile)
        except FileNotFoundError:
            pass

    def backup(self, file=None, force: bool = False, full: bool = False) -> str:
        """
        Backs up the Dataset and any attached results/extra files to zip file

        Args:
            file:
                target path
            force:
                overwrite file if it exists
            full:
                only collects runner results if False (defaults ``False``)
        Returns:
            path to zip file
        """
        if file is None:
            file = f"{self.name}.zip"

        if not file.endswith(".zip"):
            raise ValueError(f'backup file "{file}" must be of .zip type')

        file = os.path.abspath(file)

        if os.path.isfile(file) and not force:
            raise RuntimeError(
                f'backup file "{file}" exists, use a different name or '
                f"force=True to overwrite"
            )

        logger.debug("writing to file %s", file)

        try:
            with open(Dataset._manifest_file, "w+", encoding="utf-8") as o:
                o.write(self.dbfile)

            with ZipFile(file, "w") as z:
                logger.info("storing file %s", self.database.path)
                z.write(self.database.path)
                z.write(Dataset._manifest_file)

                pwd = os.getcwd()
                for rfile in self.collect_files(
                    remote_check=False, results_only=not full
                ):
                    if not os.path.isfile(rfile):
                        continue
                    logger.info("storing file %s", rfile)
                    rfile = rfile.replace(pwd, ".")

                    z.write(rfile)
        finally:
            os.remove(Dataset._manifest_file)

        return file

    @classmethod
    def restore(cls, file, force: bool = False) -> "Dataset":
        """
        Restore from backup file `file`

        Args:
            file:
                File to restore from
            force:
                Set to True to overwrite any existing Dataset

        Returns:
            Dataset
        """
        # backup archive
        arch = ZipFile(file)
        # get the name of the Database file to recreate the Dataset
        with arch.open(name=Dataset._manifest_file, mode="r") as a:
            dbfile = a.read().decode("utf-8")
        if not force and os.path.exists(dbfile):
            raise RuntimeError(
                f"Dataset already exists, either restore with "
                f"force=True or delete the dbfile '{dbfile}'"
            )

        # extract all files that aren't the manifest
        files = [f for f in arch.namelist() if f is not Dataset._manifest_file]
        arch.extractall(members=files)
        # recreate and return
        dataset = Dataset.from_file(dbfile)

        if os.path.exists(Dataset._manifest_file):
            os.remove(Dataset._manifest_file)

        return dataset

    @property
    def runner_dict(self) -> dict:
        """
        Stored runners in dict form, where the keys are the append id
        """
        return dict(self._runs)

    @property
    def runners(self) -> list:
        """
        Stored runners as a list
        """
        return list(self.runner_dict.values())

    @property
    def states(self) -> List[RunnerState]:
        """
        Runner states as a list of RunnerState
        """
        return [runner.state for runner in self.runners]

    @property
    def string_states(self) -> List[str]:
        """
        Runner states as a list of strings
        """
        return [runner.state.state for runner in self.runners]

    @property
    def function(self) -> Union[Function, Script, None]:
        """
        Currently stored Function wrapper
        """
        return self._function

    @property
    def is_python(self):
        return isinstance(self.function, Function)

    @property
    def extra(self) -> str:
        """Returns the global level extra"""
        return self._global_run_extra

    @extra.setter
    def extra(self, extra):
        """Sets the global level extra"""
        self._global_run_extra = extra

    @property
    def shebang(self) -> str:
        """returns the url shebang"""
        return self.url.shebang

    @shebang.setter
    def shebang(self, shebang: str):
        """sets the url shebang"""
        self.url.shebang = shebang

    def _script_sub(self, avoid_nodes: bool = False, **sub_args) -> str:
        """
        Substitutes run argmuents into the computer script, if it exists

        Args:
            avoid_nodes (bool):
                ignore submission scripts if True
            **sub_args:
                jobscript arguments

        Returns:
            (str):
                jobscript
        """
        # generate a default script to be used if there's no script method
        default = [self.shebang, self._script]

        extra_cache = []
        url_extra = getattr(self.url, "extra", None)
        if try_value(url_extra) is not None:
            extra_cache.append(try_value(url_extra))

        if try_value(self.extra) is not None:
            extra_cache.append(try_value(self.extra))

        extras = ["runner_extra", "tmp_extra"]
        for key in extras:
            extra = sub_args.get(key, None)
            if extra is not None:
                extra_cache.append(extra)
        default += extra_cache
        default = "\n".join(default)
        if avoid_nodes:
            logger.info("creating a jobscript for the login nodes")
            return default
        if not self._computer:
            logger.info("not a computer, returning base script")
            return default
        if "name" not in sub_args:
            logger.info("name not found in args, appending self name %s", self.name)
            sub_args["name"] = self.name
        sub_args["extra"] = "\n".join(item for item in extra_cache if item != "")
        return self.url.script(**sub_args)

    @property
    def script(self, **sub_args) -> str:
        """
        Currently stored run script

        Args:
            sub_args:
                arguments to substitute into the script() method

        Returns:
            (str):
                arg-substituted script
        """
        sub_args.update(self.run_args)
        return self._script_sub(**sub_args)

    @script.setter
    def script(self, script: str) -> None:
        """
        Set the run script
        """
        if script == self.script:
            return
        self._script = script
        self.set_runner_states("created", force=True)
        self.verbose.print(
            "Warning! The script has changed, "
            "this will allow runners to be resubmitted!",
            level=1,
        )

    @property
    def add_newline(self) -> bool:
        """
        Returns True if add_newline is set

        This controls if scripts have an additional newline enforced at the end
        """
        return self._add_newline

    @add_newline.setter
    def add_newline(self, add_newline: bool) -> None:
        """Sets the add_newline property"""
        self._add_newline = add_newline

    @property
    def submitter(self) -> str:
        """Currently stored submission command"""
        return self.url.submitter

    @submitter.setter
    def submitter(self, submitter) -> None:
        """Set the submission command"""
        self.url.submitter = submitter

    @property
    def shell(self) -> str:
        return self.url.shell

    @shell.setter
    def shell(self, shell: str):
        self.url.shell = shell

    @property
    def url(self) -> URL:
        """
        Currently stored URL object
        """
        if not hasattr(self, "_url"):
            # noinspection PyTypeChecker
            self.url = None
        return self._url

    @url.setter
    def url(self, url: Union[URL, None] = None) -> None:
        """
        Verifies and sets the URL to be used.
        Will create an empty (local) url connection if url is None

        Args:
            url (URL):
                url to be verified
        """
        logger.info("new url is being set to %s", url)
        if url is None:
            logger.info("no URL specified for this dataset, creating localhost")
            self._url = URL(verbose=self.verbose)
        else:
            if not isinstance(url, URL):
                raise ValueError("URL is not a valid URL instance")
            self._url = url

        if not type(url) == URL and issubclass(type(url), URL):  # noqa: E721
            self._computer = True

        timeout = self.run_args.get("timeout", None)
        max_timeouts = self.run_args.get("max_timeouts", None)

        self._url.timeout = timeout
        self._url.max_timeouts = max_timeouts

        try:
            self.transport.url = self.url
        except AttributeError:
            pass

    @property
    def transport(self) -> tp.transport.Transport:
        """
        Currently stored Transport system
        """
        if getattr(self.url, "_transport", None) is None:
            self.url._set_default_transport()
        return self.url.transport

    @transport.setter
    def transport(self, transport: Union[tp.transport.Transport, None] = None) -> None:
        """
        Updates URL transport to `transport`

        Args:
            transport (Transport):
                transport to be verified
        """
        self.url.transport = transport

    @property
    def serialiser(self) -> sr.serial.serial:
        """Returns the stored serialiser object"""
        if not hasattr(self, "_serialiser"):
            self.serialiser = None
        return self._serialiser

    @serialiser.setter
    def serialiser(self, serialiser: sr.serial.serial = None) -> None:
        """
        Verifies and sets the serialiser to be used.
        Will use serialjson if serialiser is None

        Args:
            serialiser (serialiser):
                serialiser to be verified
        """
        if serialiser is None:
            logger.info("no serialiser specified, creating basic json")
            self._serialiser = sr.serialjson()

        else:
            if not isinstance(serialiser, sr.serial.serial):
                raise ValueError("serialiser is not a valid serial instance")
            self._serialiser = serialiser

    def remove_database(self) -> None:
        """Deletes the database file"""
        os.remove(self.dbfile)

    @property
    def name(self) -> str:
        """Name of this dataset"""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Sets the dataset name"""
        if not isinstance(name, str):
            try:
                raise ValueError(f"name {name} is not str type")
            except TypeError as ex:
                raise ValueError(f"name {name.__name__} is not str type") from ex

        self._name = name

    @property
    def uuid(self) -> str:
        """This Dataset's full uuid (64 characcter)"""
        return self._uuid

    @property
    def short_uuid(self) -> str:
        """This Dataset's short format (8 character) uuid"""
        return self._uuid[:8]

    def set_runner_states(
        self,
        state: str,
        uuids: list = None,
        extra: str = None,
        force: bool = False,
    ) -> None:
        """
        Update runner states to `state`

        Args:
            (str) state:
                state to set
            (list) uuids:
                list of uuids to update, updates all if not passed
        """
        logger.info("updating runner states")
        runners = []
        if uuids is not None:
            logger.info("using uuid list: %s", uuids)
            for runner in self.runners:
                if runner.uuid in uuids:
                    runners.append(runner)
        else:
            runners = self.runners

        for runner in runners:
            runner.set_state(newstate=state, force=force)

            if extra is not None:
                runner.state.extra = extra

    def check_all_runner_states(self, state: str) -> bool:
        """
        Check all runner states against `state`, returning True if `all`
        runners have this state

        Args:
            state (str):
                state to check for

        Returns (bool):
            all(states)
        """
        return all([r == state for r in self.states])

    @property
    def last_run(self) -> Union[int, None]:
        """
        Returns the unix time of the last _run call

        Returns:
            (int): unix time of last  _run call, or None if impossible
        """
        if self._last_run > 0:
            return self._last_run
        return None

    @property
    def run_summary_limit(self) -> int:
        """
        If there are more runners than this number,
        the run output will be summed up rather than printed
        """
        return self._run_summary_limit

    @run_summary_limit.setter
    def run_summary_limit(self, lim: int):
        """Sets the run summary limit"""
        self._run_summary_limit = lim

    @property
    def summary_only(self) -> bool:
        """
        Returns True if the number of runners exceeds the summary limit.
        Otherwise, returns False.

        Used for printing a shortened output when running.
        """
        return len(self.runners) > self._run_summary_limit

    def retry_failed(self, *args, **kwargs) -> None:
        """
        Retries all failed runners

        Takes args and kwargs, passes them to run
        """
        kwargs["uuids"] = [r.uuid for r in self.failed]
        kwargs["force"] = True
        self.run(*args, **kwargs)

    def stage(
        self,
        uuids: List[str] = None,
        force: bool = False,
        dependency_call: bool = False,
        extra: str = "",
        force_ignores_success: bool = False,
        verbose: Verbosity = None,
        **run_args,
    ) -> bool:
        """
        Stage all runners, generating all files and preparing for
        transfer and execution.

        Returns a boolean, True if any new content was written.
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose

        if self.dependency and not dependency_call:
            logger.info(
                "dataset %s is a part of a dependency chain, calling from there", self
            )
            return self.dependency.stage(
                force_ignores_success=force_ignores_success,
                extra=extra,
                verbose=verbose,
                **run_args,
            )
        verbose.print("Staging Dataset", level=1, end="... ")
        verbose.print("", level=2)  # newline for higher verbose levels

        run_args["force"] = force

        if uuids is not None:
            logger.info("Staging only runners %s", format_iterable(uuids))
        else:
            if all(r.state == "staged" for r in self.runners) and not force:
                verbose.print("All runners already staged, skipping", level=1)
                return False
            uuids = [r.uuid for r in self.runners]

        # first step is to handle the writing of the scripts to the local dir
        # the runners take care of this
        master_content = [
            "export sourcedir=$PWD",
            f"source $sourcedir/{self.bash_repo.name}",
        ]
        if self._fresh_dataset:
            master_content.append(f"rm -f {self.manifest_log.name}")
        # runner staging
        staged = False
        nstaged = 0
        for runner in self.runners:
            if runner.uuid not in uuids:
                logger.info(
                    "Runner %s (uuid %s) is not in selection", runner, runner.short_uuid
                )
                continue

            ready = runner.stage(
                extra=extra,
                force_ignores_success=force_ignores_success,
                verbose=verbose,
                **run_args,
            )

            self.sanitise_run_arg_paths(runner.derived_run_args)

            if ready:
                staged = True
                nstaged += 1
        # file writing
        bash_cache = []  # list of submitters to create bash functions for
        uuid_cache = []
        for runner in self.runners:
            if runner.uuid not in uuids:
                continue
            if not runner.derived_run_args.get("force", False):
                if runner.state < "staged":
                    continue
                if runner.state >= "submit pending":
                    continue

            if runner.submitter not in bash_cache:
                bash_cache.append(runner.submitter)

            runline = runner.generate_runline(remote_dir=self.remote_dir)
            master_content.append(runline)

            uuid_cache.append(runner.uuid)
            if runner.state < "staged":
                runner.set_state(newstate="staged", force=True)
            staged = True

        if not staged:
            logger.info("no runners completed the run assessment, skipping")
            verbose.print("No Runners staged", level=1)
            return staged

        # next, we need the repositories for the runners to import from
        self._write_to_repo()
        self._write_to_bash_repo(bash_cache=bash_cache)
        # write the master file
        self.master_script.write(master_content, add_newline=self.add_newline)
        verbose.print(f"Staged {nstaged}/{len(self.runners)} Runners", level=1)

        return staged

    def transfer(
        self,
        uuids: List[str] = None,
        force: bool = False,
        dependency_call: bool = False,
        extra: str = "",
        force_ignores_success: bool = False,
        verbose: Verbosity = None,
        **run_args,
    ) -> bool:
        """
        Transfer the files to the remote
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose

        if self.dependency and not dependency_call:
            logger.info(
                "dataset %s is a part of a dependency chain, calling from there", self
            )
            return self.dependency.transfer(verbose=verbose)

        staged = self.stage(
            uuids=uuids,
            extra=extra,
            force=force,
            force_ignores_success=force_ignores_success,
            verbose=verbose,
            **run_args,
        )
        uuid_cache = []
        for runner in self.runners:
            if uuids is not None and runner.uuid not in uuids:
                continue
            if not runner.derived_run_args.get("force", False):
                if runner.state >= "transferred":
                    continue

            self.transport.queue_for_push(runner.jobscript)
            if self.function is not None:
                self.transport.queue_for_push(runner.runfile)

            logger.info("queuing extra files to send")
            for file in runner.extra_files_send:
                self.transport.queue_for_push(file)

            uuid_cache.append(runner.uuid)

        if len(uuid_cache) == 0 and not staged:
            verbose.print("No Transfer required", level=1)
            return False

        verbose.print(
            f"Transferring for {len(uuid_cache)}/{len(self.runners)} Runners", level=1
        )

        self.transport.queue_for_push(self.master_script)

        self.transport.queue_for_push(self.bash_repo)
        self.transport.queue_for_push(self.repofile)

        self.master_script.chmod(755)
        self.transport.transfer(verbose=verbose)

        self.set_runner_states("transferred", uuids=uuid_cache)
        return True

    def run(
        self,
        force: bool = False,
        dry_run: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
        uuids: list = None,
        extra: str = "",
        force_ignores_success: bool = False,
        dependency_call: bool = False,
        **run_args,
    ) -> bool:
        """
        Run the functions

        Args:
            force (bool):
                force all runs to go through, ignoring checks
            dry_run (bool):
                create files, but do not run
            verbose:
                Sets local verbose level
            uuids (list):
                list of uuids to run
            extra:
                extra text to add to runner jobscripts
            failed_only (bool):
                If True, `force` will submit only failed runners
            force_ignores_success (bool):
                If True, `force` takes priority over `is_success` check
            dependency_call (bool):
                Internally used to block recursion issues with dependencies
            run_args:
                any arguments to pass to the runners during this run.
                will override any "global" arguments set at Dataset init
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose

        if self.dependency and not dependency_call:
            logger.info(
                "dataset %s is a part of a dependency chain, calling from there", self
            )
            return self.dependency.run(
                dry_run=dry_run,
                force_ignores_success=force_ignores_success,
                extra=extra,
                verbose=verbose,
                **run_args,
            )

        self.avoid_runtime()
        runtime = int(time.time())
        logger.info("#### Dataset run called at %s", runtime)
        self._run_cmd = None

        run_args["force"] = force

        if os.name == "nt" and self.url.is_local:
            raise RuntimeError(LOCALWINERROR)

        if uuids is not None:
            logger.info("running only runners %s", format_iterable(uuids))

        self.transfer(
            uuids=uuids,
            extra=extra,
            force_ignores_success=force_ignores_success,
            verbose=verbose,
            **run_args,
        )

        uuid_cache = []
        asynchronous = False
        for runner in self.runners:
            if uuids is not None and runner.uuid not in uuids:
                continue
            if not runner.derived_run_args.get("force", False):
                if runner.state >= "submit pending":
                    continue

            if runner.derived_run_args["asynchronous"]:
                asynchronous = True

            uuid_cache.append(runner.uuid)

        if len(uuid_cache) == 0:
            return self._run_finalise(verbose=verbose, passthrough=False)

        verbose.print(
            f"Remotely executing {len(uuid_cache)}/{len(self.runners)} Runners", level=1
        )

        launch_cmd = self.shell
        cmd = f"cd {self.remote_dir} && {launch_cmd} {self.master_script.name}"
        if not dry_run:
            extra = None if not force else "forced"
            self.set_runner_states(
                state="submit pending", uuids=uuid_cache, extra=extra
            )
            self._run_cmd = self.url.cmd(cmd, asynchronous=asynchronous)
            self._fresh_dataset = False
        else:
            self.set_runner_states("dry run", uuids=uuid_cache)
            msg = f"launch command: {cmd}"
            logger.info(msg)
            verbose.print(msg, 1)
            self._run_cmd = self.url.cmd(
                cmd, asynchronous=asynchronous, verbose=verbose, dry_run=True
            )

        self._last_run = runtime
        return self._run_finalise(verbose=verbose, passthrough=not dry_run)

    def _write_to_repo(self, skip_function: bool = False) -> None:
        """
        Write out the repo file for this run

        Args:
            skip_function: Skip function dump if True. Used for dependencies.
        """
        repo_class_file = repo.__file__
        with open(repo_class_file, encoding="utf8") as o:
            base_file = o.read()
        # add manifest filename, and convert to list for appending
        base_file = [base_file.replace("{manifest_filename}", self.manifest_log.name)]
        self.repofile.write(base_file)

        content = []
        # no function to write if we are relying on the scripts
        if self.is_python:
            content += [
                "\n### serialiser functions ###",
                self.serialiser.dumpfunc(),
                self.serialiser.loadfunc(),
            ]
            # allow dependencies to skip the function and handle it themselves
            if not skip_function:
                try:
                    content += [
                        "\n### primary function ###",
                        self.function.source,
                    ]
                except AttributeError:
                    pass

        if len(cached_functions) != 0:
            content += [
                "\n### cached functions ###",
                *[f.source for f in cached_functions.values()],
            ]

        if len(content) == 0:
            return

        self.repofile.append("\n".join(content))

    def _write_to_bash_repo(self, bash_cache: list):
        content = []
        for i, sub in enumerate(bash_cache):
            add_doc = i == len(bash_cache) - 1
            script_run = not self.is_python
            run_func = create_run_function(
                submitter=sub,
                script_run=script_run,
                add_docstring=add_doc,  # add docstring for first function only
                manifest_filename=self.manifest_log.name,
            )
            content.append(run_func)

        self.bash_repo.write(content)

    def _run_finalise(
        self,
        passthrough: bool,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> bool:
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        if self.summary_only:
            run = 0
            skip = 0
            force = 0
            for runner in self.runners:
                if runner._run_state == "skip":
                    skip += 1
                elif runner._run_state == "run":
                    run += 1
                elif runner._run_state == "force":
                    run += 1
                    force += 1
            # condense the summary into a single line
            output = [f"Of {len(self.runners)} runners, {run} run,"]

            if force > 0:
                output.append(f"({force} forced)")
            if skip > 0:
                output.append(f"({skip} skipped)")

            output.append("see run_log for more info")

            msg = " ".join(output)
            logger.info(msg)
            verbose.print(msg, level=1)

        self.update_db()
        return passthrough

    @property
    def run_cmd(self) -> CMD:
        """
        Access to the storage of CMD objects used to run the scripts

        Returns:
            (list): List of CMD objects
        """
        return self._run_cmd

    def check_states(self, state: str) -> dict:
        """Call the repo "last_time" method remotely"""
        cmd = (
            f"cd {self.remote_dir} && "
            f"{self.url.python} {self.repofile.name} "
            f"None check_last {state}"
        )

        string = self.url.cmd(cmd).stdout
        result = json.loads(string)

        return result

    def check_started(self) -> dict:
        """Check when runners started remotely, using the manifest"""
        return self.check_states("started")

    @property
    def is_finished(self) -> list:
        """Queries the finished state of this Dataset"""
        return self._is_finished()

    @property
    def is_finished_force(self) -> list:
        """Queries the finished state of this Dataset"""
        return self._is_finished(force=True)

    def _is_finished(
        self,
        check_dependency: bool = True,
        dependency_call: bool = False,
        force: bool = False,
    ) -> list:
        """
        Query the runners and return their states

        Args:
            check_dependency: Checks the dependency if True
            dependency_call: Internal flag to prevent recursion errors
            force: passthrough for update_runners force
        """
        self.avoid_runtime()
        t = int(time.time())
        logger.info("#### _is_finished called at %s", t)
        fin = {r.uuid: r.is_finished for r in self.runners}

        if all(r.is_success for r in self.runners):
            logger.info("all runners are marked Succeeded, returning early")
            return list(fin.values())

        self.update_runners()

        if (
            self.run_cmd is not None
            and self.run_cmd.returncode is not None
            and self.run_cmd.is_finished
            and not self.run_cmd.succeeded
        ):
            stderr = self.run_cmd.communicate(ignore_errors=True)["stderr"]
            msg = f"Dataset encountered an issue:\n{stderr}"
            if force:
                warnings.warn(msg)
            else:
                raise RuntimeError(msg)

        if check_dependency and not dependency_call and self.dependency is not None:
            self.dependency.check_failure()

        fin = {r.uuid: r.is_finished for r in self.runners}

        return list(fin.values())

    @property
    def all_finished(self) -> bool:
        """
        Check if `all` runners have finished

        Returns (bool):
            True if all runners have completed their runs
        """
        return all(self.is_finished)

    @property
    def all_success(self) -> bool:
        """Returns True if all runners report that they have succeeded"""
        self._is_finished()
        return all([r.is_success for r in self.runners])

    def wait(
        self,
        interval: Union[int, float] = 10,
        timeout: Union[int, float] = None,
        watch: bool = False,
        success_only: bool = False,
        only_runner: Runner = None,
        force: bool = False,
    ) -> None:
        """
        Watch the calculation, printing updates as runners complete

        Args:
            interval:
                check interval time in seconds
            timeout:
                maximum time to wait in seconds
            watch:
                print an updating table of runner states
            success_only:
                Completion search ignores failed runs if True
            only_runner:
                wait for only this runner to complete
            force:
                Raises dataset level errors as errors if True

        Returns:
            None
        """

        def wait_condition():
            states = self._is_finished(force=force)

            if only_runner is not None:
                return only_runner.is_finished
            if success_only:
                return self.all_success

            return all([s for s in states if s is not None])

        def print_status():
            from IPython.display import clear_output

            # noinspection PyUnboundLocalVariable
            clear_output(wait=True)
            print(f"watching {len(self.runners)} runners, with a {interval}s interval")

            if timeout:
                print(f"will time out if t > {timeout}")

            print(f"t={dt:.1f}")

            for runner in self.runners:
                statetxt = runner.state
                print(f"{runner.name}, {statetxt}")

        t0 = int(time.time())
        # check all non None states
        while not wait_condition():
            dt = int(time.time()) - t0

            if watch:
                print_status()

            if timeout is not None and dt > timeout:
                raise RuntimeError("wait timed out")

            time.sleep(interval)

        if watch:
            print_status()

    def fetch_results(
        self,
        results: bool = True,
        errors: bool = True,
        extras: bool = True,
        force: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
    ):
        """
        Fetch results from the remote, and store them in the runner results property

        Args:
            results:
                fetch result files
            errors:
                fetch error files
            extras:
                fetch extra files
        Returns:
            None
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        self.avoid_runtime()
        t = int(time.time())
        logger.info("#### fetch_results called at %s", t)

        # if we're going to rely on runner states, we should update them
        self._is_finished(check_dependency=False, force=force)

        transfer = False
        level = 3
        verbose.print("Fetching results", level=1)
        verbose.print("Checking Runner states", level=level)
        for runner in self.runners:
            verbose.print(f"\t{runner}", level=level, end="... ")
            if runner.state == "satisfied":
                verbose.print("Already marked Satisfied", level=level, end=", ")
                if not runner.verify_local_files():
                    logger.info(
                        "runner resultfile is missing locally, attempting a pull"
                    )
                    self.transport.queue_for_pull(runner.resultfile)
                    for file in runner.extra_files_recv:
                        self.transport.queue_for_pull(file)
                    transfer = True
                    verbose.print("with missing files", level=level, end="")
                else:
                    logger.info("runner is satisfied, no work needed")
                    verbose.print("no work needed", level=level, end="")

            elif runner.state == "completed":
                transfer = True
                verbose.print("Completed", level=level, end="")
                if results:
                    logger.info("runner marked as completed, pulling result")
                    self.transport.queue_for_pull(runner.resultfile)
                    verbose.print(", pulling result", level=level, end="")
                else:
                    logger.info("runner marked as completed, but ignoring result")
                    verbose.print("ignoring result", level=level, end="")

                if runner.errorfile.size != 0:
                    # There can be an error and result, so we should pull both
                    if errors:
                        verbose.print(", pulling error", level=level, end="")
                        self.transport.queue_for_pull(runner.errorfile)

                if extras:
                    verbose.print(", pulling extras", level=level, end="")
                    for file in runner.extra_files_recv:
                        self.transport.queue_for_pull(file)

            elif runner.state == "failed":
                transfer = True
                logger.info("runner marked as completed, pulling error")
                verbose.print(", runner marked as failed", level=level, end="")
                if errors:
                    self.transport.queue_for_pull(runner.errorfile)
                    verbose.print(", pulling error", level=level, end="")

            verbose.print("", level=level)

        if transfer:
            logger.info("a transfer was requested, transfer and read")
            self.transport.transfer(raise_errors=False, verbose=verbose)

            self.update_runners()

            for runner in self.runners:
                runner.read_local_files()

            for cmd in self.transport.cmds:
                if cmd.stderr:
                    warnings.warn(
                        f"\nWARNING! When transferring files, "
                        f"fetch_results encountered an error:\n{cmd.stderr}"
                    )
        else:
            verbose.print("No Transfer Required", level=1)

        self.update_db()

    def update_runners(
        self,
        runners: Union[list, None] = None,
        dependency_call: bool = False,
    ):
        """
        Collects the manifest file, updating runners

        Args:
            runners: list of runners to update, usually used for dependencies
            dependency_call: internal flag to avoid dependecy loops
        """
        if self.dependency is not None and not dependency_call:
            self.dependency.update_runners()

        if runners is None:
            runners = self.runners

        logger.debug("parsing log")
        manifest = repo.Manifest(instance_uuid=self.short_uuid)
        manifest.runner_mode = True

        log_collection = self.url.cmd(
            f"cat {self.manifest_log.remote}", raise_errors=False
        )
        if log_collection.succeeded:
            runner_data = manifest.parse_log(string=log_collection.stdout)

            for runner in runners:
                if runner.short_uuid in runner_data["log"]:
                    logger.debug("Runner %s:", runner)
                    for logline in runner_data["log"][runner.short_uuid]:
                        timestamp, data = logline
                        logger.debug("\t%s", logline)
                        runner.set_state(
                            newstate=data.strip(),
                            state_time=timestamp,
                            check_state=False,
                        )
                else:
                    logger.debug("Runner %s not found", runner)

    @property
    def results(self) -> list:
        """
        Access the results of the runners

        Returns (list):
            ``runner.result`` for each runner
        """
        self.avoid_runtime()
        logger.info("#### Dataset results called")
        # check first for errors
        n_errors = len([e for e in self.errors if e is not None])
        if n_errors != 0:
            msg = (
                f"Warning! Found {n_errors} error(s), also check the `errors` property!"
            )
            logger.warning(msg)
            self.verbose.print(msg, level=1)
        return [r.result for r in self.runners]

    @property
    def errors(self) -> list:
        """
        Access the errors of the runners

        Returns (list):
            ``runner.error`` for each runner
        """
        self.avoid_runtime()
        logger.info("#### Dataset errors called")
        return [r.error for r in self.runners]

    @property
    def failed(self) -> list:
        """
        Returns a list of failed runners

        Returns:
            list of failed runners
        """
        return [r for r in self.runners if r.state.failed]

    def prepare_for_transfer(self) -> None:
        """Ensures that the Transport class is able to function"""
        # ensure transport url is synced
        self.transport.url = self.url

    def avoid_runtime(self) -> None:
        """
        Call for last_runtime sensitive operations such as is_finished and fetch_results

        Waits for 1s if we're too close to the saved _last_run time

        Returns:
            None
        """
        self.prepare_for_transfer()
        checktime = int(time.time())

        if checktime <= self._last_run:
            logger.info("call is too soon after last run, sleeping for 1s")
            time.sleep(1)


def line_starts_with_uuid(line: str) -> bool:
    """
    Checks if line starts with a short uuid

    returns True if line starts like "a1b2c3d4", False otherwise
    """
    search = re.compile(r"[0-9A-F]{8}", re.IGNORECASE)

    return re.match(search, line) is not None



"""
URL base class for connecting to remote systems
"""

import copy
import logging
import os
import time
from collections import deque
from typing import Optional, Union

import remotemanager.transport as tp
from remotemanager.connection.cmd import CMD
from remotemanager.connection.testing_object import ConnectionTest
from remotemanager.logging_utils.decorate_verbose import make_verbose
from remotemanager.logging_utils.utils import format_iterable
from remotemanager.logging_utils.verbosity import Verbosity
from remotemanager.utils import ensure_list, random_string
from remotemanager.utils.uuid import generate_uuid

logger = logging.getLogger(__name__)


@make_verbose
class URL:
    """
    Container to store the url info for a Remote run

    The url should contain everything pertaining to the _remote_, allowing
    Dataset to be remote-agnostic

    Arguments:
        host (str):
            host address of the remote system
        user (str):
            username for the remote system
        port (int, str):
            port to connect to for ssh tunnels
        verbose (bool):
            base-verbosity for connections
        timeout (int):
            time to wait before issuing a timeout for cmd calls
        max_timeouts (int):
            number of times to attempt cmd communication in case of a timeout
        python (str):
            string used to initiate a python instance
        raise_errors (bool):
            set false to ignore errors by default in cmd calls
        passfile (str):
            absolute path to password file for sshpass calls
        envpass (str):
            environment variable containing absolute path to password file for
        sshpass_override (str):
            override the sshpass string
            sshpass calls
        cmd_history_depth (int):
            number of cmd calls to store history for
        landing_dir (str):
            set the directory which is treated as the ssh endpoint
        ssh_insert (str):
            any extra flags you wish to add to the ssh call
        quiet_ssh (bool):
            Option to add -q flag to ssh calls. This suppresses errors and warnings
            that machines can sometimes put on stderr, breaking runs. Defaults True
        kwargs:
            any extra args that may end up here from a Dataset or Computer are
            discarded
    """

    _localhost = "localhost"

    _do_not_package = ["_urlutils"]

    _submitter_default = "bash"

    def __init__(
        self,
        host: Optional[str] = None,
        user: Optional[str] = None,
        port: Optional[int] = None,
        verbose: Optional[Union[int, bool, "Verbosity"]] = 1,
        timeout: Optional[int] = 5,
        max_timeouts: Optional[int] = 3,
        python: Optional[str] = "python",
        submitter: Optional[str] = _submitter_default,
        shell: Optional[str] = "bash",
        raise_errors: Optional[bool] = True,
        keyfile: Optional[str] = None,
        passfile: Optional[str] = None,
        envpass: Optional[str] = None,
        sshpass_override: Optional[str] = None,
        cmd_history_depth: Optional[int] = 10,
        landing_dir: Optional[str] = None,
        ssh_insert: Optional[str] = "",
        ssh_prepend: Optional[str] = None,
        ssh_override: Optional[str] = None,
        quiet_ssh: Optional[bool] = True,
        shebang: Optional[str] = "#!/bin/bash",
        transport: Optional[tp.transport.Transport] = None,
        **kwargs,
    ):
        self.verbose = verbose

        if host is None:
            if user is not None:
                raise ValueError(
                    f"user is set to {user}, but host is unset, "
                    f'did you mean to set host="{user}"?'
                )

            host = URL._localhost
        elif "@" in host and user is None:
            user, host = host.split("@")

        self._conn = {"user": user, "host": host, "port": port}

        self.timeout = timeout
        self.max_timeouts = max_timeouts
        self._submitter = None
        self._shell = None
        self.submitter = submitter
        self.shell = shell

        self.python = python
        self.shebang = shebang

        self._home = None
        self._landing_override = landing_dir

        self._cmd_history_depth = cmd_history_depth
        self._cmd_history = deque(maxlen=self._cmd_history_depth)
        # explicit path takes precedent over environment variable
        if passfile is None and envpass is not None:
            passfile = os.environ[envpass]

        if "ignore_errors" in kwargs:
            raise_errors = not kwargs.pop("ignore_errors")

        self._keyfile = keyfile
        self._passfile = passfile
        self._passfile_override = sshpass_override

        self._raise_errors = raise_errors
        self._ssh_override = ssh_override
        self._ssh_prepend = ssh_prepend
        self._ssh_insert = ssh_insert
        self.quiet_ssh = quiet_ssh

        msg = f"new url created with url details:{format_iterable(self._conn)}"
        logger.info(msg)
        self.verbose.print(msg, level=2)

        self._callcount = 0

        self._latency = 0
        self._connection_test = None

        self._transport = self._validate_transport(transport)

    def __deepcopy__(self, memo: dict):
        """
        Prevents a failure when copying a URL that has cmds in the history

        Taken from this answer https://stackoverflow.com/a/15774013
        """
        self.reset_cmd_history()

        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    @property
    def uuid(self):
        return generate_uuid(self._conn)

    @property
    def short_uuid(self):
        return self.uuid[:8]

    @property
    def raise_errors(self) -> bool:
        return self._raise_errors

    @raise_errors.setter
    def raise_errors(self, r: bool):
        self._raise_errors = r

    @property
    def ignore_errors(self) -> bool:
        return not self.raise_errors

    @ignore_errors.setter
    def ignore_errors(self, ignore: bool):
        self._raise_errors = not ignore

    @property
    def call_count(self) -> int:
        return self._callcount

    @property
    def user(self) -> str:
        """
        Currently configured username
        """
        return self._conn["user"]

    @user.setter
    def user(self, user):
        """
        Set the user attribute
        """
        self._conn["user"] = user

    @property
    def host(self) -> str:
        """
        Currently configured hostname
        """
        return self._conn["host"] or URL._localhost

    @host.setter
    def host(self, host):
        """
        Set the host attribute
        """
        self._conn["host"] = host

    @property
    def port(self) -> int:
        """
        Currently configured port (defaults to 22)
        """
        port = self._conn["port"] or 22
        return port

    @port.setter
    def port(self, port):
        """
        Set the host attribute
        """
        self._conn["port"] = port

    @property
    def userhost(self) -> str:
        """
        `user@host` string if possible, just `host` if user is not present
        """
        if self.user is None:
            return self.host
        else:
            return f"{self.user}@{self.host}"

    @property
    def ssh_insert(self) -> str:
        return self._ssh_insert

    @ssh_insert.setter
    def ssh_insert(self, val: str):
        self._ssh_insert = val

    @property
    def ssh_prepend(self) -> str:
        return self._ssh_prepend

    @ssh_prepend.setter
    def ssh_prepend(self, val: str):
        self._ssh_prepend = val

    @property
    def submitter(self):
        if self._submitter is not None:
            return self._submitter
        logger.debug("submitter is None but shell is set, returning shell")
        return self._shell

    @submitter.setter
    def submitter(self, submitter):
        self._submitter = submitter

    @property
    def shell(self):
        return self._shell

    @shell.setter
    def shell(self, shell):
        if self.submitter == self._submitter_default:
            self.submitter = shell
        self._shell = shell

    @property
    def home(self) -> str:
        if self._home is None:
            self.gethome()
        return self._home

    def gethome(self) -> str:
        self._home = self.expandvars("$HOME")
        return self._home

    def clearhome(self):
        """
        Clear the home property

        Returns:
            None
        """
        self._home = None

    @property
    def landing_dir(self):
        if self._landing_override is not None:
            return self._landing_override

        logger.debug("fallback to returning $HOME")
        return "$HOME"

    @landing_dir.setter
    def landing_dir(self, landing):
        logger.info("updating landing dir to %s", landing)
        self._landing_override = landing

    @property
    def transport(self):
        if self._transport is None:
            self._set_default_transport()
        return self._transport

    def _validate_transport(
        self, transport: Union[None, tp.transport.Transport]
    ) -> Union[None, tp.transport.Transport]:
        if transport is None:
            return None
        if not isinstance(transport, tp.transport.Transport):
            raise ValueError(
                f"{transport} is not a valid transport instance ({type(transport)})"
            )
        return transport

    @transport.setter
    def transport(self, transport: Union[None, tp.transport.Transport]):
        if transport is None:
            return self._set_default_transport()

        transport = self._validate_transport(transport)
        self._transport = transport
        self._transport.set_remote(self)

    def _set_default_transport(self):
        logger.debug("Internal transport is None, creating basic rsync")
        self._transport = tp.rsync(url=self, verbose=self.verbose)

    @property
    def ssh(self) -> str:
        """
        ssh insert for commands on this connection
        """
        if self.is_local:
            return ""

        if self._ssh_override:
            return self._ssh_override

        ret = []

        if self.passfile is not None:
            ret.append(self.passfile)

        if self.ssh_insert != "":
            ret.append(f"ssh {self.ssh_insert} -p {self.port}")
        else:
            ret.append(f"ssh -p {self.port}")

        if self.quiet_ssh:
            ret.append("-q")

        if self._keyfile:
            ret.append(f"-i {self.keyfile}")

        ret.append(self.userhost)

        return " ".join(ret)

    @ssh.setter
    def ssh(self, newssh: str) -> None:
        """
        Allows forced override of the ssh command

        Inserting extra flags into the ssh can be done as follows:

        >>> url = URL()
        >>> print(url.ssh)
        >>> "ssh"
        >>> url.ssh = "LANG=C " + url.ssh
        >>> print(url.ssh)
        >>> "LANG=C ssh"

        Args:
            newssh (str):
                new ssh string to insert

        Returns:
            None
        """
        self._ssh_override = newssh

    def clear_ssh_override(self):
        """
        Wipe any override applied to ssh. Can also be done by setting
        url.ssh = None

        Returns:
            None
        """
        self._ssh_override = None

    @property
    def keyfile(self):
        if self._keyfile is None:
            return
        p = self._keyfile.replace("~", os.environ["HOME"])
        if not os.path.isfile(p):
            raise RuntimeError(f"could not find ssh key file at {self._keyfile}")

        return self._keyfile

    @keyfile.setter
    def keyfile(self, file):
        self._keyfile = file

    @property
    def sshpass_override(self) -> str:
        return self._passfile_override

    @sshpass_override.setter
    def sshpass_override(self, override: str):
        self._passfile_override = override

    @property
    def passfile(self) -> [str, None]:
        """
        Returns the sshpass string

        .. versionadded:: 0.10.2
            No longer returns just the file, to facilitate overriding

        Returns:
            sshpass full string
        """
        if self._passfile_override is not None:
            return self._passfile_override

        if self._passfile is None:
            return

        p = self._passfile.replace("~", os.environ["HOME"])
        if not os.path.isfile(p):
            raise RuntimeError(f"could not find password file at {self._passfile}")

        return f"sshpass -f {self._passfile}"

    @passfile.setter
    def passfile(self, file):
        self._passfile = file

    def tunnel(
        self,
        local_port: int,
        remote_port: int,
        local_address: str = None,
        background: bool = False,
        verbose: Union[None, int, float, Verbosity] = None,
        dry_run: bool = False,
    ) -> CMD:
        """
        Create a tunnel to the host, between the local and remote ports

        Args:
            local_port (int):
                port to open on the local side
            remote_port (int):
                port to access on the  remote side
            local_address (str):
                change the local address of the tunnel
            background (bool):
                creates the tunnel asynchronously if True (default False)
            verbose:
                override verbose setting for this call
            dry_run (bool):
                do not execute if True> Defaults to False

        Returns:
            CMD:
                The CMD instance responsible for the tunnel
        """

        def validate_port(p: int) -> int:
            if not isinstance(p, int):
                raise ValueError(f"Port {p} is not int type")
            if p <= 1024:
                raise ValueError("Ports below 1024 are reserved for root")
            return p

        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        local_port = validate_port(local_port)
        remote_port = validate_port(remote_port)

        local_address = "" if local_address is None else local_address
        if not local_address.endswith(":"):
            local_address += ":"

        cmd = (
            f"{self.ssh} -N -L {local_address}{local_port}"
            f":{self.host}:{remote_port} {self.host}"
        )

        t = self.cmd(
            cmd, local=True, dry_run=dry_run, timeout=False, asynchronous=background
        )
        if not dry_run:
            verbose.print(
                f"Created a tunnel to host {self.host} with the command:\n"
                f"{cmd}\n"
                f"Tunnel PID: {t.pid}",
                level=1,
            )

        return t

    @property
    def is_local(self):
        """
        True if this connection is purely local
        """
        host = self.host
        if host == URL._localhost:
            return True
        elif host.startswith("127."):
            return True
        return False

    def ping(
        self,
        n: int = 5,
        timeout: int = 30,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> float:
        """
        Perform and monitor a ping command

        Args:
            n (int): number of pings to aim for
            timeout (int): kill the process and return 0 if
                this period is elapsed

        Returns:
            (float) latency in ms
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        msg = f"pinging {self.host}"
        logger.info(msg)
        verbose.print(msg, level=2)

        tmpfile = f"ping_{random_string()}"

        def cleanup(process):
            process.kill(verbose=verbose)
            try:
                os.remove(tmpfile)
                logger.debug("removed temporary file")
            except FileNotFoundError:
                logger.debug("temporary ping file could not be found")

        ping = self.cmd(
            f"ping {self.host}", local=True, stdout=tmpfile, asynchronous=True
        )

        lines = []
        t0 = time.time()
        while len(lines) < n + 1:
            with open(tmpfile, mode="r", encoding="utf8") as o:
                lines = o.readlines()

            time.sleep(0.5)

            if time.time() - t0 > timeout:
                msg = f"ping timed out after {timeout} seconds"
                logger.warning(msg)
                verbose.print(msg, 1)
                cleanup(ping)
                return -1

        cleanup(ping)

        times = []
        for line in lines[1:]:
            times.append(self._process_ping_line(line))

        avg = sum(times) / n
        msg = f"ping times: {times} -> {avg}"
        logger.info(msg)
        verbose.print(msg, level=2)
        return avg

    def _process_ping_line(self, line: str) -> float:
        """
        Function to parse the contents of a line output from a ping command

        Args:
            line (str):
                ping command line
        Returns:
            (float): ping time in seconds
        """
        timing = line.split("time=")[1].strip()  # get timing
        val, units = timing.split()
        logger.debug("parsed line %s -> %s (%s)", line, val, units)

        if units != "ms":
            raise ValueError(f"Unknown ping units {units}")

        return float(val) * 1e-3  # convert to seconds

    def expandvars(self, string: str) -> str:
        """
        'echo' a string on the remote, returning the result

        Args:
            string:
                string to be expanded

        Returns:
            str
        """
        wrapped = f'bash -c "echo {string}"'

        return self.cmd(wrapped).stdout

    def cmd(
        self,
        cmd: str,
        asynchronous: bool = False,
        local: Union[bool, None] = None,
        stdout: str = None,
        stderr: str = None,
        timeout: int = None,
        max_timeouts: int = None,
        raise_errors: bool = None,
        dry_run: bool = False,
        prepend: bool = False,
        force_file: bool = False,
        landing_dir: str = None,
        stream: bool = False,
        verbose: Union[None, int, bool, Verbosity] = None,
    ) -> CMD:
        """
        Creates and executes a command

        Args:
            asynchronous (bool):
                run this command asynchronously
            cmd (str):
                command to execute
            local (bool, None):
                force a local or remote execution. Defaults to None
            stdout (str):
                optional file to redirect stdout to
            stderr (str):
                optional file to redirect stderr to
            timeout (int):
                time to wait before issuing a timeout
            max_timeouts (int):
                number of times to attempt communication in case of a timeout
            raise_errors (bool):
                override for global setting. Raise any stderr if encountered
            dry_run (bool):
                don't exec the command if True, just returns the string
            prepend (bool):
                always attempt to use ssh_prepend
            force_file (bool):
                passthrough for CMD force_file argument
            landing_dir:
                set the directory which is treated as the ssh endpoint
            stream:
                Attempts to stream the output as it arrives on stdout

        Returns (CMD):
            returned command instance
        """
        if verbose is not None:
            verbose = Verbosity(verbose)
        else:
            verbose = self.verbose
        _remote_call = False
        if self._landing_override is not None or landing_dir is not None:
            ld = landing_dir or self.landing_dir
            logger.debug("modifying cmd with landing dir %s", ld)
            cmd = f"cd {ld} && {cmd}"

        if local is not None and not local:
            logger.info("forced remote call; appending ssh to cmd")
            cmd = f"{self.ssh} '{cmd}'"
            _remote_call = True
        elif local is None and not self.is_local:
            logger.info("implicit remote call; appending ssh to cmd")
            cmd = f"{self.ssh} '{cmd}'"
            _remote_call = True
        if raise_errors is None:
            raise_errors = self._raise_errors

        timeout = self.timeout if timeout is None else timeout
        max_timeouts = self.max_timeouts if max_timeouts is None else max_timeouts

        if self.ssh_prepend is not None:
            logger.info("found prepend string %s", self.ssh_prepend)
            prep = self.ssh_prepend.strip()

            if prepend:
                logger.debug("force adding prepend string %s", prep)
                cmd = f"{prep} {cmd}"
            elif _remote_call:
                logger.debug("implicit adding prepend string %s", prep)
                cmd = f"{prep} {cmd}"

        thiscmd = CMD(
            cmd.strip(),
            asynchronous=asynchronous,
            stdout=stdout,
            stderr=stderr,
            timeout=timeout,
            max_timeouts=max_timeouts,
            raise_errors=raise_errors,
            force_file=force_file,
            stream=stream,
            verbose=verbose,
        )

        if dry_run:
            return thiscmd

        thiscmd.exec()
        if not local:
            self._callcount += 1

        self._cmd_history.append(thiscmd)

        return thiscmd

    @property
    def cmd_history(self):
        return self._cmd_history

    @property
    def cmd_history_depth(self):
        return self._cmd_history_depth

    def reset_cmd_history(self):
        self._cmd_history = deque(maxlen=self._cmd_history_depth)

    @cmd_history_depth.setter
    def cmd_history_depth(self, newdepth: int):
        """
        Updates the history depth, and creates a new dequeue populated with as many of
        the existing cmds as possible

        Args:
            newdepth (int):
                new depth to capture
        Returns:

        """
        logger.info(
            "updating history depth from %s to %s", self.cmd_history_depth, newdepth
        )
        self._cmd_history_depth = newdepth

        newqueue = deque(maxlen=self.cmd_history_depth)

        for item in self.cmd_history:
            newqueue.append(item)

        self._cmd_history = newqueue

    @property
    def utils(self):
        """
        Handle for the URLUtils module
        """
        if getattr(self, "_urlutils", None) is None:
            self._urlutils = URLUtils(self)
        return self._urlutils

    def test_connection(self) -> None:
        """
        Create a ConnectionTest instance and run the tests

        Returns:
            None
        """
        testing_object = ConnectionTest(self)
        testing_object.exec()

        self._connection_test = testing_object

    @property
    def connection_test(self) -> Union[ConnectionTest, None]:
        """
        Return the connection test object

        Returns:
            ConnectionTest: testing object
        """
        return self._connection_test

    @property
    def connection_data(self) -> dict:
        """
        Returns the results of a previous connection test

        Returns:
            (dict) connection data
        """
        if self.connection_test is not None:
            return self.connection_test.data
        return {}

    @property
    def latency(self) -> [float, None]:
        """
        Attempts to access the latency property of the stored ConnectionTest

        Returns:
            (float): connection latency in seconds, if available. Else None
        """
        if self._connection_test is not None:
            return self._connection_test.latency
        return None

    def script(self, **kwargs):
        raise NotImplementedError

    @staticmethod
    def download_file(file_url: str, filename: str) -> None:
        """
        Download file at url `file_url` and write the content out to `filename`

        Args:
            file_url: url of file
            filename: name to write content to
        """
        import requests

        response = requests.get(file_url)

        if response.status_code == requests.codes.ok:
            # Save the file
            fld, file = os.path.split(filename)
            if fld != "" and not os.path.exists(fld):
                os.makedirs(fld)

            with open(filename, "wb") as file:
                file.write(response.content)
            print(f"Grabbed file '{filename}'")
        else:
            raise RuntimeError(f"Could not find a file at: {file_url}")


class URLUtils:
    """
    Extra functions to go with the URL class, called via URL.utils

    As it requires a parent `URL` to function, and is instantiated with a
    `URL`, there is little to no purpose to using this class exclusively

    Arguments:
        parent (URL):
            parent class to provide utils to
    """

    def __init__(self, parent: URL):
        logger = logging.getLogger(__name__ + ".URLUtils")
        logger.info("creating a utils extension to parent: %s", parent)

        self._parent = parent

    def file_mtime(
        self,
        files: list,
        local: bool = None,
        python: bool = False,
        ignore_empty: bool = False,
        dry_run: bool = False,
    ) -> dict:
        """
        Check file modification times of [files]

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            python (bool):
                ensure python style search is used
            ignore_empty (bool):
                also check the filesize, ignoring empty files
            dry_run (bool):
                print command only

        Returns (dict):
            {file: mtime (unix)} dictionary
        """

        logger.info("performing stat on files: %s", files)
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)
        times, error = self._file_mtime(files, local, python, dry_run)

        if dry_run:
            # in this instance "times" is simply the command
            return times

        logger.info("received:")
        logger.info(times)
        logger.info(error)
        output = {}
        for file in files:
            if file in times:
                mtime = times[file][0]
                fsize = times[file][1]
                if ignore_empty and fsize == 0:
                    output[file] = None
                else:
                    output[file] = mtime
            else:
                output[file] = None

        return output

    def _file_mtime(self, files: list, local: bool, python: bool, dry_run: bool):
        """
        Perform the "stat -c %Y" command on a list of files,
        returning the result. Uses a python command backup if this fails

        Args:
            files (list):
                list of files to check
            local (bool):
                force a local search
            python (bool):
                force the python override
            dry_run (bool):
                print command only

        Returns:
            (list): list of file unix times
        """
        sep = ","

        def stat():
            logger.debug("attempting raw stat command on files")
            basecmd = f"stat -c %n{sep}%Y{sep}%s"
            if len(files) == 1:
                cmd = f"{basecmd} {files[0]}"
            else:
                cmd = f"{basecmd} {{" + ",".join(files) + "}"

            ret = self._parent.cmd(
                cmd, local=local, raise_errors=False, dry_run=dry_run
            )

            if dry_run:
                return ret, "", "", ""

            times = {}
            for line in ret.stdout.split("\n"):
                try:
                    fname = line.split(sep)[0]
                    mtime = int(float(line.split(sep)[1]))
                    fsize = int(float(line.split(sep)[2]))

                    times[fname] = (mtime, fsize)
                except IndexError:
                    pass

            return times, ret.stderr.split("\n"), ret.returncode, ret.stderr

        def pystat():
            logger.debug("attempting python stat on files")
            ex = f"""import os
files={files}
for f in files:
\ttry: print(f'{{f}}{sep}{{os.stat(f).st_mtime}}{sep}{{os.stat(f).st_size}}')
\texcept FileNotFoundError: print(f)"""

            cmd = f'{self._parent.python} -c "{ex}"'

            ret = self._parent.cmd(
                cmd, local=local, raise_errors=False, dry_run=dry_run
            )

            if dry_run:
                return ret, None

            times = {}
            error = []
            for line in ret.stdout.split("\n"):
                try:
                    fname = line.split(sep)[0]
                    mtime = int(float(line.split(sep)[1]))
                    fsize = int(float(line.split(sep)[2]))

                    times[fname] = (mtime, fsize)
                except IndexError:
                    error.append(line.strip())

            return times, error

        files = ensure_list(files)

        if not python:
            t, e, returncode, stderr = stat()
            if returncode in [126, 127] or "illegal option" in stderr:
                logger.warning("stat failed, falling back on python")
                return pystat()

            return t, e

        return pystat()

    def file_presence(
        self, files: list, local: bool = None, dry_run: bool = False
    ) -> dict:
        """
        Search for a list of files, returning a boolean presence dict

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        logger.info("checking for presence of files: %s", files)
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)

        times = self.file_mtime(files, local=local, dry_run=dry_run)

        if dry_run:
            return times

        return {f: times[f] is not None for f in files}

    def search_folder(
        self, files: list, folder: str, local: bool = None, dry_run: bool = False
    ) -> dict:
        """
        Search `folder` for `files`, returning a boolean presence dict

        Arguments:
            files (list):
                list of filenames to check for. Optionally, a string for a
                single file
            folder (str):
                folder to scan
            local (bool):
                perform the scan locally (or remotely)
            dry_run (bool):
                print command only

        Returns (dict):
            {file: present} dictionary
        """
        if local is None:
            local = self._parent.is_local
        fpath = os.path.abspath(folder) if local else folder

        logger.debug("scanning folder %s", fpath)
        logger.debug("searching for files:")
        logger.debug("%s", format_iterable(files))

        ls_return = self.ls(fpath, local=local, as_list=True, dry_run=dry_run)

        if dry_run:
            logger.info("dry run, returning command")
            return ls_return

        scan = [os.path.basename(f) for f in ls_return]

        logger.debug("scan sees:")
        logger.debug("%s", format_iterable(scan))

        if isinstance(files, str):
            logger.info("files is a string, running in singular mode")
            ret = {files: os.path.basename(files) in scan}
        else:
            ret = {file: os.path.basename(file) in scan for file in files}

        return ret

    def touch(
        self,
        file: str,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> CMD:
        """
        perform unix `touch`, creating or updating `file`

        Arguments:
            file (str):
                filename or path to file
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        logger.debug("utils touch on file %s", file)
        fname = os.path.abspath(file) if local else file
        return self._parent.cmd(
            f"touch {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

    def mkdir(
        self,
        file: str,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> CMD:
        """
        perform unix `mkdir -p`, creating a folder structure

        Arguments:
            file (str):
                name or path to folder
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        logger.debug("utils mkdir on path %s", file)
        fname = os.path.abspath(file) if local else file
        return self._parent.cmd(
            f"mkdir -p {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

    def ls(
        self,
        file: str,
        as_list: bool = True,
        local: bool = None,
        raise_errors: bool = None,
        dry_run: bool = False,
    ) -> [CMD, list]:
        """
        Identify the files present on the directory

        Arguments:
            file (str):
                name or path to folder.
            as_list (bool):
                convert to a list format
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered
            dry_run (bool):
                print command only

        Returns (CMD, list):
            CMD instance for the command, or the list if as_list is True
        """
        if local is None:
            local = self._parent.is_local
        logger.debug("utils ls on path %s", file)
        fname = os.path.abspath(file) if local else file

        ret = self._parent.cmd(
            f"ls {fname}", local=local, raise_errors=raise_errors, dry_run=dry_run
        )

        if as_list and not dry_run:
            ret = [f for f in ret.stdout.split("\n") if f != ""]
        return ret



import os
import re
import time
from datetime import datetime, timezone
from typing import Union

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import dir_delta


class TrackedFile(SendableMixin):
    __slots__ = [
        "_remote_path",
        "_local_path",
        "_file",
        "_last_seen",
        "_size",
        "_binary",
    ]

    def __init__(
        self,
        local_path: str,
        remote_path: str,
        file: str,
        binary: bool = False,
    ):
        self._remote_path = remote_path
        self._local_path = local_path
        self._file = file

        self._binary = binary

        self._last_seen = {"remote": -1, "local": -1}
        self._size = -1

    def __repr__(self) -> str:
        return f"{{{self.name}}}"

    def __fspath__(self) -> str:
        return self.local

    @property
    def binary(self) -> bool:
        """Access to private _binary attribute."""
        if hasattr(self, "_binary"):
            return self._binary
        return False

    @binary.setter
    def binary(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("binary attribute must be bool.")
        self._binary = value

    @property
    def name(self) -> str:
        """Returns the filename"""
        return self._file

    @property
    def importstr(self) -> str:
        """
        Returns the filename without extension

        Suitable for python imports
        """
        return os.path.splitext(self._file)[0]

    @property
    def remote(self) -> str:
        """Returns the full remote path"""
        return os.path.join(self._remote_path, self.name)

    @property
    def local(self) -> str:
        """Returns the full local path"""
        return os.path.join(self._local_path, self.name)

    @property
    def remote_dir(self) -> str:
        """Returns the remote dir"""
        return self._remote_path

    @property
    def local_dir(self) -> str:
        """Returns the full local dir"""
        return self._local_path

    def relative_remote_path(self, other: str) -> str:
        """
        Return a path relative to `cwd`

        Args:
            other:
                working dir to compare against

        Returns:
            relative path
        """
        # if our remote path is an abspath, we already have what we need
        if os.path.isabs(self.remote_dir):
            return self.remote

        # we're already in the remote, just return the filename
        if self.remote_dir == other:
            return self.name

        # find the deepest shared path, treat it as a "root"
        stem = os.path.commonpath([self.remote_dir, other])
        # find how far down this stem is from `other`
        dirdelta = dir_delta(stem, other)
        # generate a ../ string that steps "down" to the common path
        down = "../" * dirdelta

        tmp_remote = self.remote_dir.replace(stem, "").strip("/")
        # rebuild up from our virtual root
        return os.path.join(down, tmp_remote, self.name)

    @property
    def content(self) -> Union[str, None]:
        """
        Attempts to read the file contents

        Returns None if the file cannot be read
        """
        return self._read()

    def _read(self) -> Union[str, None]:
        """
        Attempts to read the file contents
        """
        if not os.path.isfile(self.local):
            return None

        if self.binary:
            mode_b = "b"
            encoding = None
        else:
            mode_b = ""
            encoding = "utf8"

        mode = f"r{mode_b}+"
        with open(file=self.local, mode=mode, encoding=encoding) as o:
            self.confirm_local()
            return o.read()

    def _write(
        self, content: Union[str, list], append: bool, add_newline: bool
    ) -> None:
        """
        Write to the file

        Args:
            content:
                Content to add
            append:
                Appends to file if True, overwrites otherwise
            add_newline:
                Finish the write with an extra newline if True
        """
        if isinstance(content, bytes):
            self.binary = True

        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        # try to join lists, falling back on a basic str coercion
        if not isinstance(content, str) and not self.binary:
            try:
                content = "\n".join(content)
            except TypeError:
                content = str(content)

        if append:
            mode_a = "a"
        else:
            mode_a = "w"

        if self.binary:
            mode_b = "b"
            encoding = None
        else:
            mode_b = ""
            encoding = "utf8"

        mode = f"{mode_a}{mode_b}+"
        with open(file=self.local, mode=mode, encoding=encoding) as o:
            o.write(content)

            if not self.binary and add_newline and not content.endswith("\n"):
                o.write("\n")

        self.confirm_local()

    def write(self, content: Union[str, list], add_newline: bool = True) -> None:
        """
        Write `content` to the local copy of the file

        Args:
            content:
                content to write
            add_newline:
                enforces a newline character at the end of the write if True
                (default True)
        """
        self._write(content, append=False, add_newline=add_newline)

    def append(self, content: Union[str, list], add_newline: bool = True) -> None:
        """
        Append `content` to the local copy of the file

        Args:
            content:
                content to append
            add_newline:
                enforces a newline character at the end of the write if True
                (default True)
        """
        self._write(content, append=True, add_newline=add_newline)

    def confirm_local(self, t: Union[int, None] = None) -> None:
        """
        Confirm sighting of the file locally

        Args:
            t: Optionally set the time to `t` instead of time.time()
        """
        if t is None:
            t = int(datetime.now(timezone.utc).replace(tzinfo=None).timestamp())
        self._last_seen["local"] = t

    def confirm_remote(self, t: Union[int, None] = None) -> None:
        """
        Confirm sighting of the file on the remote

        Args:
            t: Optionally set the time to `t` instead of time.time()
        """
        if t is None:
            t = int(time.time())
        self._last_seen["remote"] = t

    @property
    def exists_local(self) -> bool:
        """Returns True if the file exists locally"""
        return os.path.exists(self.local)

    def last_seen(self, where: str) -> int:
        """
        Returns the last_seen_<where>

        Where <where> is remote or local

        Args:
            where:
                remote or local
        """
        return self._last_seen[where]

    @property
    def last_seen_local(self) -> int:
        """Returns the time this file was last confirmed seen on the local machine"""
        return self.last_seen("local")

    @property
    def last_seen_remote(self) -> int:
        """Returns the time this file was last confirmed seen on the remote machine"""
        return self.last_seen("remote")

    @property
    def local_mtime(self) -> int:
        """Returns the mtime of the local file"""
        if self.exists_local:
            self.confirm_local()

            return int(os.path.getmtime(self.local))
        return -1

    @property
    def size(self) -> int:
        """
        Returns the filesize (needs to be set externally)

        -1 if not set
        """
        return self._size

    @size.setter
    def size(self, size: int) -> None:
        """Sets the filesize"""
        self._size = size

    def sub(self, source, target, mode: str = "python") -> bool:
        """
        Substitute source for target

        Args:
            source: file content to sub
            target: intended replacement
            mode: function to use, default to Python replace

        returns:
            True if a substitution was executed, False otherwise
        """
        content = self.content

        if content is None:
            return False

        mode = mode.lower()
        if mode == "python":
            content = self.content.replace(source, target)
        elif mode == "regex":
            content = re.sub(source, target, content)
        else:
            raise ValueError(f"mode {mode} not understood")

        self.write(content)
        return True

    def chmod(self, mod: int):
        """
        chmod the local file, if it exists

        python3 chmod requires an octal input, so convert the base10 input
        """
        if self.exists_local:
            if not isinstance(mod, int):
                raise ValueError(f"chmod {mod} must be int-type")

            convert = f"0o{mod}"
            if len(convert) > 6:
                raise ValueError(f"Converted octal {convert} is too long")

            mod = int(oct(eval(convert)), base=8)  # pylint: disable=eval-used
            os.chmod(self.local, mod)



import functools
import logging
import operator
import os.path
import shutil
import time
import warnings
from typing import Union, Any

import yaml

from remotemanager.utils import get_version, ensure_filetype, recursive_dict_update
from remotemanager.utils.version import Version

logger = logging.getLogger(__name__)

# this list marks points where datasets cannot be carried through
barriers = {
    ("<=", "0.11.x"): ("==", "0.12.x"),  # engine difference
    ("==", "0.12.x"): (">=", "0.13.x"),  # engine difference
}


class Database:
    """
    Database file handler for use in the Dataset.

    .. warning::
        Interacting with this object directly could cause unstable behaviour.
        It is best to allow Dataset to handle the Database

    Args:
        file (str):
            filename to write to
    """

    _versionkey = "~database-version~"

    def __init__(self, file):
        file = ensure_filetype(file, "yaml")

        self._path = file
        self._storage = self._read()
        self._tree = None

    def _read(self) -> dict:
        """
        Wipes the tree cache and reads from the database file, creating it if
        it does not already exist.
        """
        self._tree = None  # wipe tree cache
        logger.info("reading %s", self._path)
        try:
            with open(self._path, "r") as o:
                data = yaml.safe_load(o)
        except FileNotFoundError:
            logger.warning("file not found, creating")
            data = {Database._versionkey: get_version()}
            self._write(data)
        check_version = True
        try:
            version = data.pop(Database._versionkey)
        except KeyError:
            check_version = False
            version = "0.0.0"

        logger.info("database file version: %s", version)

        oldversion = Version(version)
        currentversion = Version(get_version())
        if check_version and oldversion < currentversion:
            msg = (
                f"database file (version {version}) is less than "
                f"current remotemanager install version ({currentversion})"
            )
            warnings.warn(msg)
            logger.warning(msg)

            if not check_version_barriers(oldversion, currentversion, barriers):
                backup_file = self.backup()

                msg = (
                    f"{oldversion} cannot be upgraded to {currentversion}. "
                    f"\nThe dataset was backed up locally to {backup_file}, it is "
                    f"recommended that you downgrade before continuing. This can "
                    f"be done with:\n`pip install remotemanager=={oldversion}`"
                    "\nIf you haven't already, consider creating a virtual "
                    "environment.\nTo ignore this warning and continue, remove the "
                    f"`~database-version~: {oldversion}` line from the bottom of "
                    f"the database file."
                )

                logger.error(msg)
                raise RuntimeError(msg) from None
        # url storage is blocked as of 0.13.4, strip it
        uuid = self._get_first_key(data)
        if uuid is not None and data[uuid].get("_url", None) is not None:
            del data[uuid]["_url"]

        return data

    def _write(self, data: dict):
        """
        Clears the tree cache and writes to the file

        Args:
            data (dict):
                writes `data` to the database file
        """
        self._tree = None  # wipe tree cache

        data[Database._versionkey] = get_version()

        tmp_path = self._path + ".tmp"
        with open(tmp_path, "w+") as o:
            yaml.dump(data=data, stream=o, sort_keys=False)

        timeout = 5
        t0 = time.time()
        while not os.path.exists(tmp_path):
            # if the file isn't ready yet, the move will raise a strange error
            # waiting can help with this
            time.sleep(0.1)

            if time.time() > t0 + timeout:
                raise RuntimeError(
                    f"Database file {tmp_path} was never created "
                    f"after {timeout}s. This is usually caused by "
                    f"a naming clash, if you have concurrent Datasets, "
                    f"consider giving them differing names."
                )

        shutil.move(tmp_path, self._path)

    def read(self):
        """
        Read the database file
        """
        self._storage = self._read()

    def write(self):
        """
        Write the database to file
        """
        self._write(self._storage)

    def update(self, payload):
        """
        Update the database with payload (dict)

        Args:
            payload (dict):
                Dictionary to recursively update with. Usually called with the
                output of object.pack()
        """

        logger.info("updating stored info")

        self._storage = recursive_dict_update(self._storage, payload)
        self.write()

    def _get_first_key(self, d: dict) -> Any:
        if len(d) == 0:
            return None
        firstkey = next(iter(d))

        data_uuid = d.get(firstkey, {}).get("_uuid", None)
        if data_uuid != firstkey:
            return None

        return firstkey

    @property
    def stored_uuid(self) -> Union[str, None]:
        return self._get_first_key(self._storage)

    @property
    def path(self):
        """
        Path to current database file
        """
        return self._path

    @property
    def tree(self) -> list:
        """
        Returns a list of path-like strings for the stored database dict

        Returns (list):
            List of path-like strings
        """
        if self._tree is not None:
            return self._tree
        self._tree = self.climb(self._storage)
        return self._tree

    def climb(self, data: dict, branch: list = None) -> list:
        """
        "climb" a dictionary tree, returning a list of path-like strings for
        each element

        .. note::
            This method is intended for use within the `tree` property, and
            could cause unintended behaviour if called directly, though will
            allow you to produce a `tree` like list for any dictionary. Use
            with caution.

        Args:
            data (dict):
                dictionary to treat
            branch (list):
                current branch, used for recursion

        Returns (list):
            list of path-like strings
        """

        if branch is None:
            branch = []

        joinchar = "/"

        ret = []

        if len(data) == 0:
            ret.append(joinchar.join(branch))

        for key in data:
            tmp = []
            if isinstance(data[key], dict):
                tmp.append(key)
                ret += self.climb(data[key], branch + tmp)
            else:
                ret.append(joinchar.join(branch + [key]))
        return ret

    def find(self, key: str, greedy: bool = False) -> dict:
        """
        Find the first instance of key within the database tree

        Args:
            key:
                key (uuid) to look for
            greedy:
                returns the first value found if True. Faster, but does not respect
                the tree "layers"

        Returns:
            database tree below key
        """

        def get_minimum_path(p: str, k: str) -> list:
            """
            Returns the path segments of path p up to target k

            >>> get_minimum_path('a/b/c/d/', 'b')
            >>> ['a', 'b']

            Args:
                p:
                    /path/containing/target/k/value
                k:
                    target within path p

            Returns:
                list of steps within path, leading to k
            """
            fullpath = p.split("/")
            return fullpath[: fullpath.index(k) + 1]

        target = None
        for path in self.tree:
            if key in path:
                tmp = get_minimum_path(path, key)
                if greedy:
                    target = tmp
                    break
                # we should prefer targets lower down the tree
                if target is None or len(tmp) < len(target):
                    target = tmp

        if target is None:
            return {}

        data = chain_get(self._storage, target)

        return data

    def backup(self, file=None) -> str:
        """
        Back up the database file to `file`, defaults to `{self.path}.bk`

        Args:
            file:
                backup file target location

        Returns:
            backup file path
        """
        if file is None:
            file = f"{self.path}.bk"

        shutil.copyfile(self.path, file)

        return file


def chain_get(d: dict, keys: list):
    """
    Get item from a nested dict using a list of keys

    Args:
        d (dict):
            nested dict to query
        keys (list):
            list of keys to use

    Returns:
        item from nested dict `d` at [list, of, keys]
    """
    return functools.reduce(operator.getitem, keys, d)


def check_version_barriers(old: Version, new: Version, barriers: dict) -> bool:
    """
    Takes two versions, and a dict of {old: new} that contain breaking changes

    Returns True if the two versions are compatible, False otherwise

    Args:
        old:
            old version
        new:
            new version
        barriers:
            dictionary of incompatible {old: new} parings.

            Uses the syntax of Version.match
    """
    dispatch = {
        "==": operator.eq,
        "!=": operator.ne,
        "<": operator.lt,
        ">": operator.gt,
        "<=": operator.le,
        ">=": operator.ge,
    }
    for a, b in barriers.items():
        a_op, a_ver = a
        b_op, b_ver = b

        if dispatch[a_op](old, a_ver) and dispatch[b_op](new, b_ver):
            return False

    return True
